unit sdlgfx;
{$MODE FPC}
{
	GearHead: Arena, a roguelike mecha CRPG
	Copyright (C) 2005 Joseph Hewitt

	This library is free software; you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation; either version 2.1 of the License, or (at
	your option) any later version.

	The full text of the LGPL can be found in license.txt.

	This library is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
	General Public License for more details. 

	You should have received a copy of the GNU Lesser General Public License
	along with this library; if not, write to the Free Software Foundation,
	Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
}

interface

uses
{$IFDEF PATCH_GH}
{ "sysutils" has to come before others. }
	sysutils,
	dos,
	gears_base,
	ui4gh,
	SDL,SDL_TTF,SDL_Image,texutil
{$ELSE PATCH_GH}
  {$IF DEFINED(PATCH_I18N) or DEFINED(WITH_TENC)}
	sysutils,
  {$ENDIF}
	SDL,SDL_TTF,SDL_Image,texutil,gears,dos,ui4gh
{$ENDIF PATCH_GH}
{$IFDEF PATCH_AUDIO}
	,SDL_Mixer
{$ENDIF PATCH_AUDIO}
{$IFDEF PATCH_I18N}
	,strings
{$ENDIF PATCH_I18N}
	;

Type
	SensibleSpritePtr = ^SensibleSprite;
	SensibleSprite = Record
		Name,Color: String;
		W,H: Integer;	{ Width and Height of each cell. }
{$IFDEF PATCH_GH}
		alpha_blending: Boolean;
{$ENDIF PATCH_GH}
		Img: PSDL_Surface;
		Next: SensibleSpritePtr;
	end;

	RedrawProcedureType = Procedure;

{$IFDEF PATCH_GH}
	DynamicRectPtr = ^DynamicRect;
{$ENDIF PATCH_GH}
	DynamicRect = Object
		dx,dy,w,h,anchor: Integer;
{$IFDEF PATCH_GH}
		dx_org,dy_org,w_org,h_org: Integer;
{$ENDIF PATCH_GH}
		function GetRect: TSDL_Rect;
	end;


const
{$IFDEF PATCH_GH}
	{ Moved into ui4gh.pp }
{$ELSE PATCH_GH}
	StdBlack: TSDL_Color =		( r:  0; g:  0; b:  0 );
	StdWhite: TSDL_Color =		( r:255; g:255; b:255 );
	MenuItem: TSDL_Color =		( r:90; g:156; b:179 );
	MenuSelect: TSDL_Color =	( r:128; g:255; b:230 );
	PlayerBlue: TSDL_Color =	( r:  0; g:141; b:211 );
	AllyPurple: TSDL_Color =	( r:236; g:  0; b:211 );
	EnemyRed: TSDL_Color =		( r:230; g:  0; b:  0 );
	NeutralGrey: TSDL_Color =	( r:150; g:150; b:150 );
	DarkGrey: TSDL_Color =		( r:100; g:100; b:100 );
	InfoGreen: TSDL_Color =		( r: 50; g:200; b:  0 );
	InfoHiLight: TSDL_Color =	( r:100; g:250; b:  0 );

	TextboxGrey: TSDL_Color =	( r:130; g:120; b:125 );
	NeutralBrown: TSDL_Color =	( r:240; g:201; b: 20 );
	BorderBlue: TSDL_Color =	( r:  0; g:101; b:151 );
	BrightYellow: TSDL_Color =	( r:255; g:201; b:  0 );
{$ENDIF PATCH_GH}

	ScreenWidth = 800;
	ScreenHeight = 600;
{$IFDEF PATCH_I18N}
{$ELSE PATCH_I18N}
	BigFontSize = 13;
	SmallFontSize = 11;
{$ENDIF PATCH_I18N}
	Right_Column_Width = 220;
	Dialog_Area_Height = 110;

	ANC_upperleft = 0;
	ANC_upper = 1;
	ANC_upperright = 2;
	ANC_left = 3;
	ANC_middle = 4;
	ANC_right = 5;
	ANC_lowerleft = 6;
	ANC_lower = 7;
	ANC_lowerright = 8;

	ZONE_Map: TSDL_Rect = ( x:10; y:10; w: ScreenWidth - Right_Column_Width - 30 ; h: ScreenHeight - Dialog_Area_Height - 20 );
	ZONE_Clock: TSDL_Rect = ( x: ScreenWidth - Right_Column_Width - 10 ; y:ScreenHeight - Dialog_Area_Height - 30; w:Right_Column_Width; h:20 );
	ZONE_PCInfo: TSDL_Rect = ( x:  ScreenWidth - Right_Column_Width - 10 ; y:10; w:Right_Column_Width; h:150 );
	ZONE_Dialog: TSDL_Rect = ( x:10; y: ScreenHeight - Dialog_Area_Height ; w: ScreenWidth - 20 ; h:Dialog_Area_Height-10 );

	ZONE_TitleScreenMenu: DynamicRect = ( dx:-100; dy:50; w:200; h:100; anchor: ANC_middle );
	ZONE_TitleScreenLogo: DynamicRect = ( dx:-250; dy:-190; w:500; h:215; anchor: ANC_middle );
{$IFDEF PATCH_I18N}
	ZONE_TitleScreenVersion: DynamicRect = ( dx:400; dy:-25; w:600; h:20; anchor: ANC_lowerleft );
{$ELSE PATCH_I18N}
	ZONE_TitleScreenVersion: DynamicRect = ( dx:-50; dy:-25; w:100; h:20; anchor: ANC_lowerright );
{$ENDIF PATCH_I18N}

	ZONE_TargetInfo: DynamicRect = ( dx:  -Right_Column_Width -10 ; dy:10; w:Right_Column_Width; h:150; anchor: ANC_upperright );
	ZONE_TargetDistance: DynamicRect = ( dx:  -Right_Column_Width -10 ; dy:176; w:Right_Column_Width; h:20; anchor: ANC_upperright );

	ZONE_CharGenChar: DynamicRect = ( dx:-368; dy:-210; w: 500 ; h: 420; anchor: ANC_middle );
	ZONE_CharGenMenu: DynamicRect = ( dx:148; dy:-50; w:220; h:230; anchor: ANC_middle );
	ZONE_CharGenCaption: DynamicRect = ( dx:148; dy:190; w:220; h:20; anchor: ANC_middle );
	ZONE_CharGenDesc: DynamicRect = ( dx:148; dy:-210; w:220; h:150; anchor: ANC_middle );
	ZONE_CharGenPrompt: DynamicRect = ( dx:-150; dy:-245; w:300; h:20; anchor: ANC_middle );
	ZONE_CharGenHint: DynamicRect = ( dx:-160; dy:225; w:320; h:20; anchor: ANC_middle );

	ZONE_CharViewChar: DynamicRect = ( dx:-368; dy:-260; w: 500 ; h: 420; anchor: ANC_middle );
	ZONE_CharViewMenu: DynamicRect = ( dx:148; dy:-100; w:220; h:230; anchor: ANC_middle );
	ZONE_CharViewCaption: DynamicRect = ( dx:148; dy:140; w:220; h:20; anchor: ANC_middle );
	ZONE_CharViewDesc: DynamicRect = ( dx:148; dy:-260; w:220; h:150; anchor: ANC_middle );


{$IFDEF PATCH_GH}
	ZONE_HQPilots: DynamicRect = ( dx:20; dy:10; w:200; h:400; anchor: ANC_upperleft );
	ZONE_HQMecha: DynamicRect = ( dx:240; dy:10; w:200; h:400; anchor: ANC_upperleft );
	ZONE_HQMenu: DynamicRect = ( dx:500; dy:10; w:200; h:400; anchor: ANC_upperleft );
	ZONE_HQMenu2: DynamicRect = ( dx:20; dy:430; w:420; h:80; anchor: ANC_upperleft );
	ZONE_HQMenu3: DynamicRect = ( dx:500; dy:430; w:200; h:80; anchor: ANC_upperleft );
{$ENDIF PATCH_GH}

{$IFDEF PATCH_GH}
	ZONE_TextInputPrompt: DynamicRect = ( dx:-210; dy:-70; w:420; h:52; anchor: ANC_middle );
	ZONE_TextInput: DynamicRect = ( dx:-210; dy:-12; w:420; h:16; anchor: ANC_middle );
	ZONE_TextInputBigBox: DynamicRect = ( dx:-220; dy:-73; w:440; h:80; anchor: ANC_middle );
{$ELSE PATCH_GH}
	ZONE_TextInputPrompt: DynamicRect = ( dx:-210; dy:-51; w:420; h:16; anchor: ANC_middle );
	ZONE_TextInput: DynamicRect = ( dx:-210; dy:-27; w:420; h:16; anchor: ANC_middle );
	ZONE_TextInputBigBox: DynamicRect = ( dx:-220; dy:-61; w:440; h:56; anchor: ANC_middle );
{$ENDIF PATCH_GH}
	ZONE_PhoneInstructions: DynamicRect = ( dx:-200; dy:15; w:400; h:16; anchor: ANC_middle );

	ZONE_InteractStatus: DynamicRect = ( dx:-250; dy: -210; w: 395; h: 40; anchor: ANC_middle );
	ZONE_InteractMsg: DynamicRect = ( dx: -250; dy:-120; w:395; h: 110; anchor: ANC_middle );
	ZONE_InteractMenu: DynamicRect = ( dx: -250; dy:-5; w:500; h: 120; anchor: ANC_middle );
	ZONE_InteractPhoto: DynamicRect = ( dx: 150; dy: -185; w: 100; h: 150; anchor: ANC_middle );
	ZONE_InteractInfo: DynamicRect = ( dx: -250; dy:-165; w:395; h:40; anchor: ANC_middle );
	ZONE_InteractTotal: DynamicRect = ( dx: -255; dy: -215; w: 510; h: 335; anchor: ANC_middle );
{$IFDEF PATCH_GH}
	ZONE_InteractPhotoFrame: DynamicRect = ( dx: 250; dy: -215; w: 100; h: 335; anchor: ANC_middle );
{$ENDIF PATCH_GH}


	ZONE_Menu: DynamicRect = ( dx: 10; dy:10; w:Right_Column_Width; h:205; anchor: ANC_upperleft );
	ZONE_Menu1: DynamicRect = ( dx: 10; dy:10; w:Right_Column_Width; h:100; anchor: ANC_upperleft );
	ZONE_Menu2: DynamicRect = ( dx: 10; dy:126; w:Right_Column_Width; h:80; anchor: ANC_upperleft );
{$IFDEF PATCH_GH}
	ZONE_GameOptionMenu: DynamicRect = ( dx: -10 -140; dy:10; w:140; h:205; anchor: ANC_upperright );
	ZONE_MenuRight: DynamicRect = ( dx: -10 -Right_Column_Width; dy:10; w:Right_Column_Width; h:205; anchor: ANC_upperright );
	ZONE_Menu11Right: DynamicRect = ( dx: -10 -Right_Column_Width; dy:10; w:Right_Column_Width; h:32; anchor: ANC_upperright );
	ZONE_Menu12Right: DynamicRect = ( dx: -10 -Right_Column_Width; dy:42; w:Right_Column_Width; h:64; anchor: ANC_upperright );
	ZONE_Menu1Right: DynamicRect = ( dx: -10 -Right_Column_Width; dy:10; w:Right_Column_Width; h:100; anchor: ANC_upperright );
	ZONE_Menu2Right: DynamicRect = ( dx: -10 -Right_Column_Width; dy:126; w:Right_Column_Width; h:80; anchor: ANC_upperright );
{$ENDIF PATCH_GH}

	ZONE_MemoText: DynamicRect = ( dx:-175; dy:-150; w:350; h:200; anchor: ANC_middle );
	ZONE_MemoMenu: DynamicRect = ( dx:-175; dy:55; w:350; h:50; anchor: ANC_middle );
	ZONE_MemoTotal: DynamicRect = ( dx:-180; dy:-155; w:360; h:265; anchor: ANC_middle );

	ZONE_CenterMenu: DynamicRect = ( dx:-120; dy:-155; w:240; h:210; anchor: ANC_middle );

	ZONE_FHQTitle: DynamicRect = ( dx:-165; dy:-255; w:300; h:20; anchor: ANC_middle );
	ZONE_FHQMenu: DynamicRect = ( dx:-280; dy:-210; w:292; h:340; anchor: ANC_middle );
	ZONE_FHQInfo: DynamicRect = (dx:30; dY:-210; W: 250; H: 340; anchor: ANC_middle);
	ZONE_FHQMenu1: DynamicRect = ( dx:-280; dy:-210; w:292; h:180; anchor: ANC_middle );
	ZONE_FHQMenu2: DynamicRect = ( dx:-280; dy: -15; w:292; h:145; anchor: ANC_middle );

	ZONE_BPTotal: DynamicRect = (dx:-285; dY:-215; W: 570; H: 350; anchor: ANC_middle);
	ZONE_BPHeader: DynamicRect = (dx:-280; dY:-210; W: 292; H: 40; anchor: ANC_middle);
	ZONE_EqpMenu: DynamicRect = ( dx:-280; dy:-165; w:292; h:100; anchor: ANC_middle );
	ZONE_InvMenu: DynamicRect = ( dx:-280; dy:-60; w:292; h:145; anchor: ANC_middle );
	ZONE_BPInstructions: DynamicRect = (dx:-280; dY:90; W: 292; H: 40; anchor: ANC_middle);
	ZONE_BPInfo: DynamicRect = (dx:30; dY:-210; W: 250; H: 340; anchor: ANC_middle);

	ZONE_ShopNPCName: DynamicRect = ( dx:-330; dy: -230; w: 100; h: 32; anchor: ANC_middle );
	ZONE_ShopNPCPortrait: DynamicRect = ( dx:-330; dy: -210; w: 100; h: 150; anchor: ANC_middle );
	ZONE_ShopText: DynamicRect = ( dx:-225; dy: -230; w: 287; h: 170; anchor: ANC_middle );
	ZONE_ShopPCName: DynamicRect = ( dx:-330; dy: -30; w: 100; h: 32; anchor: ANC_middle );
	ZONE_ShopPCPortrait: DynamicRect = ( dx:-330; dy: -10; w: 100; h: 150; anchor: ANC_middle );
	ZONE_ShopMenu: DynamicRect = ( dx:-225; dy: -30; w: 287; h: 170; anchor: ANC_middle );

	ZONE_ShopInfo: DynamicRect = (dx:85; dY:-225; W: 250; H: 340; anchor: ANC_middle);
	ZONE_ShopCash: DynamicRect = ( dx:135; dy: 130; w: 150; h: 16; anchor: ANC_middle );
	ZONE_ShopTop: DynamicRect = ( dx:-335; dy: -235; w: 402; h: 180; anchor: ANC_middle );
	ZONE_ShopBottom: DynamicRect = ( dx:-335; dy: -35; w: 402; h: 180; anchor: ANC_middle );

	ZONE_MoreText: DynamicRect = ( dx:-350; dy:-270; w: 700 ; h: 385; anchor: ANC_middle );
	ZONE_MorePrompt: DynamicRect = ( dx:-300; dy: 130 ; w:600; h:30; anchor: ANC_middle );

	ZONE_YesNoTotal: DynamicRect = ( dx:-180; dy:-155; w:360; h:265; anchor: ANC_middle );
	ZONE_YesNoPrompt: DynamicRect = ( dx:-175; dy:-150; w:350; h:200; anchor: ANC_middle );
	ZONE_YesNoMenu: DynamicRect = ( dx:-175; dy:55; w:350; h:50; anchor: ANC_middle );

	Console_History_Length = 240;

	GH_REPEAT_DELAY = 200;
	GH_REPEAT_INTERVAL = 50;

	{Color set constants.}
	CS_Clothing = 1;
	CS_Skin = 2;
	CS_Hair = 3;
	CS_PrimaryMecha = 4;
	CS_SecondaryMecha = 5;
	CS_Detailing = 6;


var
	Game_Screen: PSDL_Surface;
	Mouse_Pointer: PSDL_Surface;
	Game_Font,Info_Font: PTTF_Font;
	Game_Sprites,Cursor_Sprite: SensibleSpritePtr;
	Text_Messages: SAttPtr;
	Console_History: SAttPtr;
	Mouse_X, Mouse_Y: LongInt;	{ Current mouse position. }
	Animation_Phase: Integer;
	Last_Clock_Update: UInt32;

	MasterColorList: SAttPtr;

	Music_List: SAttPtr;
{	MyMusic: P_Mix_Music;}

{$IFDEF PATCH_GH_PARANOID_SAFER}
Function RndxColorString( ColorSet: Integer ): String;
{$ENDIF PATCH_GH_PARANOID_SAFER}
Function RandomColorString( ColorSet: Integer ): String;


Procedure GHFlip;

Procedure DisposeSpriteList(var LList: SensibleSpritePtr);
Procedure RemoveSprite(var LMember: SensibleSpritePtr);
Procedure CleanSpriteList;
procedure DrawSprite( Spr: SensibleSpritePtr; MyDest: TSDL_Rect; Frame: Integer );
procedure DrawAlphaSprite( Spr: SensibleSpritePtr; MyDest: TSDL_Rect; Frame: Integer );
Function ConfirmSprite( Name: String; const Color: String; W,H: Integer ): SensibleSpritePtr;

{$IFDEF PATCH_AUDIO}
Procedure PlaySoundEffect_Direct( const Name: String );
{$ENDIF PATCH_AUDIO}

{$IFDEF PATCH_I18N}
function RPGKey( var Unicode: Word; const GetStringMode: Boolean ): Char;
function RPGKey( var Unicode: Word ): Char;
{$ENDIF PATCH_I18N}
Procedure FillRectWithSprite( MyRect: TSDL_Rect; MySprite: SensibleSpritePtr; MyFrame,OffX,OffY: Integer );
Procedure FillRectWithSprite( MyRect: TSDL_Rect; MySprite: SensibleSpritePtr; MyFrame: Integer );

function RPGKey: Char;
Procedure ClrZone( var Z: TSDL_Rect );
Procedure ClrScreen;

Function PrettyPrint( msg: string; Width: Integer; var FG: TSDL_Color; DoCenter: Boolean; MyFont: PTTF_Font ): PSDL_Surface;

Procedure QuickText( const msg: String; MyDest: TSDL_Rect; Color: TSDL_Color );
Procedure QuickTinyText( const msg: String; MyDest: TSDL_Rect; Color: TSDL_Color );
Procedure CMessage( const msg: String; Z: TSDL_Rect; C: TSDL_Color );
Procedure GameMSG( const msg: string; Z: TSDL_Rect; var C: TSDL_Color );
Procedure GameMSG( const msg: string; Z: TSDL_Rect; var C: TSDL_Color; MyFont: PTTF_FONT );

{$IFDEF PATCH_GH}
{$ELSE PATCH_GH}
Function DirKey( ReDrawer: RedrawProcedureType ): Integer;
{$ENDIF PATCH_GH}
Procedure EndOfGameMoreKey;
Function TextLength( F: PTTF_Font; const msg: String ): LongInt;

Procedure RedrawConsole;
Procedure DialogMSG(msg: string); {can't const}

{$IFDEF PATCH_GH}
Function GetStringFromUser( const Prompt: String; ReDrawer: RedrawProcedureType; const Init_text: String; var canceled: Boolean ): String;
Function GetStringFromUser( const Prompt: String; ReDrawer: RedrawProcedureType; const Init_text: String ): String;
{$ENDIF PATCH_GH}
Function GetStringFromUser( const Prompt: String; ReDrawer: RedrawProcedureType ): String;
Function MsgString( const MsgLabel: String ): String;
{$IFDEF PATCH_GH}
{$ELSE PATCH_GH}
Function MoreHighFirstLine( LList: SAttPtr ): Integer;
{$ENDIF PATCH_GH}

Procedure MoreText( LList: SAttPtr; FirstLine: Integer; ReDrawer: RedrawProcedureType );

Procedure ClearExtendedBorder( Dest: TSDL_Rect );
Procedure InfoBox( MyBox: TSDL_Rect );

Procedure DrawBPBorder;
Procedure DrawCharGenBorder;
Procedure SetupCombatDisplay;
Procedure SetupHQDisplay;
Procedure SetupFactoryDisplay;
Procedure SetupYesNoDisplay;
Procedure SetupInteractDisplay( TeamColor: TSDL_Color );
Procedure SetupMemoDisplay;

{$IFDEF PATCH_I18N}
Function New_Conv_ToUni16( const arg_msg: String ): PUInt16;
Function I18N_TTF_RenderText(var font: PTTF_Font; const pline: String; var fg: TSDL_Color ): PSDL_Surface;
{$ENDIF PATCH_I18N}

Procedure SetupWizardDisplay();

implementation
{$IFDEF PATCH_GH}
uses
  {$IF DEFINED(UNIX)}
	xlib,
  {$ENDIF}
  {$IF DEFINED(WINDOWS)}
	windows,
  {$ENDIF}
	math,
	errmsg,pseudosmartpointer,
  {$IFDEF PATCH_I18N}
	i18nmsg,termenc,
  {$ENDIF PATCH_I18N}
	gears, sdlmenus
  {$IFDEF WITHOUT_SDLIM}
	,w32eb
  {$ENDIF WITHOUT_SDLIM}
	;
{$ELSE PATCH_GH}
  {$IFDEF PATCH_I18N}
uses
    {$IFDEF DEBUG}
	errmsg,
    {$ENDIF DEBUG}
	i18nmsg,termenc
    {$IFDEF WITHOUT_SDLIM}
	,w32eb
    {$ENDIF WITHOUT_SDLIM}
	;
  {$ELSE PATCH_I18N}
    {$IFDEF WITHOUT_SDLIM}
uses
      {$IFDEF DEBUG}
	errmsg,
      {$ENDIF DEBUG}
	w32eb;
    {$ELSE WITHOUT_SDLIM}
      {$IFDEF DEBUG}
uses	errmsg;
      {$ENDIF DEBUG}
    {$ENDIF WITHOUT_SDLIM}
  {$ENDIF PATCH_I18N}
{$ENDIF PATCH_GH}

{$IFDEF PATCH_AUDIO}
type
	SensibleSoundEffectListPtr = ^SensibleSoundEffectList;
	SensibleSoundEffectList = Record
		Name: String;
		Chunk: PMix_Chunk;
		Next: SensibleSoundEffectListPtr;
	end;
{$ENDIF PATCH_AUDIO}

const
	WindowName: PChar = 'GearHead Arena SDL Version';
	IconName: PChar = 'GearHead';

var
{$IFDEF PATCH_GH}
	SDL_VideoModeFlag: LongWord;
{$ENDIF PATCH_GH}
	Infobox_Border,Infobox_Backdrop: SensibleSpritePtr;
{$IFDEF PATCH_GH}
	wmi: TSDL_SysWMinfo;
  {$IF DEFINED(UNIX)}
	gX11_Lock : procedure();
	gX11_Unlock : procedure();
  {$ENDIF}
  {$IF DEFINED(WINDOWS)}
	r: Windows.RECT;
  {$ENDIF}
	FrameDrawDone: Boolean = False;
	RPGKey_LastEvent: String = '';
  {$IFDEF PATCH_I18N}
	I18N_Width_Of_One_Character: String = 'M';
  {$ENDIF PATCH_I18N}
{$ELSE PATCH_GH}
  {$IFDEF PATCH_I18N}
	I18N_Width_Of_One_Character: String = 'M';
  {$ENDIF PATCH_I18N}
{$ENDIF PATCH_GH}
{$IFDEF PATCH_AUDIO}
	SoundEffectList: SensibleSoundEffectListPtr;
{$ENDIF PATCH_AUDIO}

Function DynamicRect.GetRect: TSDL_Rect;
	{ Return the TSDL_Rect described by this DynamicRect, given the current }
	{ screen size. }
var
	MyRect: TSDL_Rect;
begin
	MyRect.W := Self.W;
	MyRect.H := Self.H;
	MyRect.X := Game_Screen^.W * (self.anchor mod 3) div 2 + Self.DX;
	MyRect.Y := Game_Screen^.H * (self.anchor div 3) div 2 + Self.DY;
	GetRect := MyRect;
end;

{$IFDEF PATCH_GH}
Procedure Recalc_DynamicRect;
	Procedure Recalc_pos( DR: DynamicRectPtr );
	begin
		DR^.dx := DR^.dx_org * ScreenSize_Width div ScreenWidth;
		DR^.dy := DR^.dy_org * ScreenSize_Height div ScreenHeight;
		DR^.w := DR^.w_org * ScreenSize_Width div ScreenWidth;
		DR^.h := DR^.h_org * ScreenSize_Height div ScreenHeight;
	end;
begin
	Recalc_pos( @ZONE_TitleScreenMenu );
	Recalc_pos( @ZONE_TitleScreenLogo );
	Recalc_pos( @ZONE_TitleScreenVersion );
	Recalc_pos( @ZONE_TargetInfo );
	Recalc_pos( @ZONE_TargetDistance );
	Recalc_pos( @ZONE_CharGenChar );
	Recalc_pos( @ZONE_CharGenMenu );
	Recalc_pos( @ZONE_CharGenCaption );
	Recalc_pos( @ZONE_CharGenDesc );
	Recalc_pos( @ZONE_CharGenPrompt );
	Recalc_pos( @ZONE_CharGenHint );
	Recalc_pos( @ZONE_CharViewChar );
	Recalc_pos( @ZONE_CharViewMenu );
	Recalc_pos( @ZONE_CharViewCaption );
	Recalc_pos( @ZONE_CharViewDesc );
	Recalc_pos( @ZONE_HQPilots );
	Recalc_pos( @ZONE_HQMecha );
	Recalc_pos( @ZONE_HQMenu );
	Recalc_pos( @ZONE_HQMenu2 );
	Recalc_pos( @ZONE_HQMenu3 );
	Recalc_pos( @ZONE_TextInputPrompt );
	Recalc_pos( @ZONE_TextInput );
	Recalc_pos( @ZONE_TextInputBigBox );
	Recalc_pos( @ZONE_PhoneInstructions );
	Recalc_pos( @ZONE_InteractStatus );
	Recalc_pos( @ZONE_InteractMsg );
	Recalc_pos( @ZONE_InteractMenu );
	Recalc_pos( @ZONE_InteractPhoto );
	Recalc_pos( @ZONE_InteractInfo );
	Recalc_pos( @ZONE_InteractTotal );
{$IFDEF PATCH_GH}
	ZONE_InteractPhotoFrame.dx := ZONE_InteractStatus.dx + ZONE_InteractStatus.w + 7;
	ZONE_InteractPhotoFrame.dy := ZONE_InteractStatus.dy;
	ZONE_InteractPhotoFrame.w := ZONE_InteractTotal.w - ZONE_InteractStatus.w - 19;
	ZONE_InteractPhotoFrame.h := ZONE_InteractMenu.dy - ZONE_InteractPhotoFrame.dy - 7;
{$ENDIF PATCH_GH}
	Recalc_pos( @ZONE_Menu );
	Recalc_pos( @ZONE_Menu1 );
	Recalc_pos( @ZONE_Menu2 );
{$IFDEF PATCH_GH}
	Recalc_pos( @ZONE_GameOptionMenu );
	Recalc_pos( @ZONE_MenuRight );
	Recalc_pos( @ZONE_Menu11Right );
	Recalc_pos( @ZONE_Menu12Right );
	Recalc_pos( @ZONE_Menu1Right );
	Recalc_pos( @ZONE_Menu2Right );
{$ENDIF PATCH_GH}
	Recalc_pos( @ZONE_MemoText );
	Recalc_pos( @ZONE_MemoMenu );
	Recalc_pos( @ZONE_MemoTotal );
	Recalc_pos( @ZONE_CenterMenu );
	Recalc_pos( @ZONE_FHQTitle );
	Recalc_pos( @ZONE_FHQMenu );
	Recalc_pos( @ZONE_FHQInfo );
	Recalc_pos( @ZONE_FHQMenu1 );
	Recalc_pos( @ZONE_FHQMenu2 );
	Recalc_pos( @ZONE_BPTotal );
	Recalc_pos( @ZONE_BPHeader );
	Recalc_pos( @ZONE_EqpMenu );
	Recalc_pos( @ZONE_InvMenu );
	Recalc_pos( @ZONE_BPInstructions );
	Recalc_pos( @ZONE_BPInfo );
	Recalc_pos( @ZONE_ShopNPCName );
	Recalc_pos( @ZONE_ShopNPCPortrait );
	Recalc_pos( @ZONE_ShopText );
	Recalc_pos( @ZONE_ShopPCName );
	Recalc_pos( @ZONE_ShopPCPortrait );
	Recalc_pos( @ZONE_ShopMenu );
	Recalc_pos( @ZONE_ShopInfo );
	Recalc_pos( @ZONE_ShopCash );
	Recalc_pos( @ZONE_ShopTop );
	Recalc_pos( @ZONE_ShopBottom );
	Recalc_pos( @ZONE_MoreText );
	Recalc_pos( @ZONE_MorePrompt );
	Recalc_pos( @ZONE_YesNoTotal );
	Recalc_pos( @ZONE_YesNoPrompt );
	Recalc_pos( @ZONE_YesNoMenu );
end;
{$ENDIF PATCH_GH}

{$IFDEF PATCH_GH_PARANOID_SAFER}
Function RndxColorString( ColorSet: Integer ): String;
	{ Select a random color string belonging to the provided color set. }
var
	C,Candidates: SAttPtr;
	it: String;
begin
	Candidates := Nil;
	C := MasterColorList;

	while C <> Nil do begin
		if ( Length( C^.Info ) > 6 ) and ( C^.Info[ ColorSet ] = '+' ) then begin
			StoreSAtt( Candidates, RetrieveAString( C^.Info ) );
		end;
		C := C^.Next;
	end;

	if Candidates <> Nil then begin
		it := SelectRndxSAtt( Candidates )^.Info;
	end else begin
		it := '100 100 100';
	end;
	DisposeSAtt( Candidates );
	RndxColorString := it;
end;
{$ENDIF PATCH_GH_PARANOID_SAFER}

Function RandomColorString( ColorSet: Integer ): String;
	{ Select a random color string belonging to the provided color set. }
var
	C,Candidates: SAttPtr;
	it: String;
begin
	Candidates := Nil;
	C := MasterColorList;

	while C <> Nil do begin
		if ( Length( C^.Info ) > 6 ) and ( C^.Info[ ColorSet ] = '+' ) then begin
			StoreSAtt( Candidates, RetrieveAString( C^.Info ) );
		end;
		C := C^.Next;
	end;

	if Candidates <> Nil then begin
		it := SelectRandomSAtt( Candidates )^.Info;
	end else begin
		it := '100 100 100';
	end;
	DisposeSAtt( Candidates );
	RandomColorString := it;
end;


Procedure GHFlip;
	{ Copy from the GH screen bitmap to the actual screen. }
var
	MyDest: TSDL_Rect;
begin
	SDL_PumpEvents;
	SDL_GetMouseState( Mouse_X , Mouse_Y );
	MyDest.X := Mouse_X;
	MyDest.Y := Mouse_Y;

	{ If a mouse pointer is defined, draw it. }
	if Mouse_Pointer <> Nil then begin
		SDL_BlitSurface( Mouse_Pointer , Nil , Game_Screen , @MyDest );
	end;

{$IFDEF PATCH_GH}
	FrameDrawDone := True;
{$ENDIF PATCH_GH}
	SDL_Flip( game_Screen );
end;

Function ScaleColorValue( V , I: Integer ): Byte;
	{ Scale a color value. }
begin
	V := ( V * I ) div 200;
	if V > 255 then V := 255;
	ScaleColorValue := V;
end;

Function MakeSwapBitmap( MyImage: PSDL_Surface; RSwap,YSwap,GSwap: PSDL_Color ): PSDL_Surface;
	{ Given a bitmap, create an 8-bit copy with pure colors. }
	{         0 : Transparent (0,0,255) }
	{   1 -  63 : Grey Scale            }
	{  64 - 127 : Pure Red              }
	{ 128 - 191 : Pure Yellow           }
	{ 192 - 255 : Pure Green            }
	{ Then, swap those colors out for the requested colors. }
var
	MyPal: Array [0..255] of TSDL_Color;
	T: Integer;
	MyImage2: PSDL_Surface;
begin
	{ Initialize the palette. }
	for t := 1 to 64 do begin
		MyPal[ T - 1 ].r := ( t * 4 ) - 1;
		MyPal[ T - 1 ].g := ( t * 4 ) - 1;
		MyPal[ T - 1 ].b := ( t * 4 ) - 1;

		MyPal[ T + 63 ].r := ( t * 4 ) - 1;
		MyPal[ T + 63 ].g := 0;
		MyPal[ T + 63 ].b := 0;

		MyPal[ T + 127 ].r := ( t * 4 ) - 1;
		MyPal[ T + 127 ].g := ( t * 4 ) - 1;
		MyPal[ T + 127 ].b := 0;

		MyPal[ T + 191 ].r := 0;
		MyPal[ T + 191 ].g := ( t * 4 ) - 1;
		MyPal[ T + 191 ].b := 0;

	end;
	MyPal[ 0 ].r := 0;
	MyPal[ 0 ].g := 0;
	MyPal[ 0 ].b := 255;

	{ Create replacement surface. }
	MyImage2 := SDL_CreateRGBSurface( SDL_SWSURFACE , MyImage^.W , MyImage^.H , 8 , 0 , 0 , 0 , 0 );
	SDL_SetPalette( MyImage2 , SDL_LOGPAL or SDL_PHYSPAL , MyPal , 0 , 256 );
	SDL_FillRect( MyImage2 , Nil , SDL_MapRGB( MyImage2^.Format , 0 , 0 , 255 ) );
	SDL_SetColorKey( MyImage2 , SDL_SRCCOLORKEY or SDL_RLEACCEL , SDL_MapRGB( MyImage2^.Format , 0 , 0, 255 ) );

	{ Blit from the original to the copy. }
	SDL_BlitSurface( MyImage , Nil , MyImage2 , Nil );

	{ Redefine the palette. }
	for t := 1 to 64 do begin
		MyPal[ T + 63 ].r := ScaleColorValue( RSwap^.R , t * 4 );
		MyPal[ T + 63 ].g := ScaleColorValue( RSwap^.G , t * 4 );
		MyPal[ T + 63 ].b := ScaleColorValue( RSwap^.B , t * 4 );

		MyPal[ T + 127 ].r := ScaleColorValue( YSwap^.R , t * 4 );
		MyPal[ T + 127 ].g := ScaleColorValue( YSwap^.G , t * 4 );
		MyPal[ T + 127 ].b := ScaleColorValue( YSwap^.B , t * 4 );

		MyPal[ T + 191 ].r := ScaleColorValue( GSwap^.R , t * 4 );
		MyPal[ T + 191 ].g := ScaleColorValue( GSwap^.G , t * 4 );
		MyPal[ T + 191 ].b := ScaleColorValue( GSwap^.B , t * 4 );
	end;
	SDL_SetPalette( MyImage2 , SDL_LOGPAL or SDL_PHYSPAL , MyPal , 0 , 256 );

	MakeSwapBitmap := MyImage2;
end;

Procedure RedefinePalette( MyImage: PSDL_Surface; RSwap,YSwap,GSwap: PSDL_Color );
	{ For a paletted image, redefine the bitmap. }
var
	MyPal: Array [0..255] of TSDL_Color;
	T: Integer;
begin
	{ Redefine the palette. }
	for t := 1 to 64 do begin
		MyPal[ T - 1 ].r := ( t * 4 ) - 1;
		MyPal[ T - 1 ].g := ( t * 4 ) - 1;
		MyPal[ T - 1 ].b := ( t * 4 ) - 1;

		MyPal[ T + 63 ].r := ScaleColorValue( RSwap^.R , t * 4 );
		MyPal[ T + 63 ].g := ScaleColorValue( RSwap^.G , t * 4 );
		MyPal[ T + 63 ].b := ScaleColorValue( RSwap^.B , t * 4 );

		MyPal[ T + 127 ].r := ScaleColorValue( YSwap^.R , t * 4 );
		MyPal[ T + 127 ].g := ScaleColorValue( YSwap^.G , t * 4 );
		MyPal[ T + 127 ].b := ScaleColorValue( YSwap^.B , t * 4 );

		MyPal[ T + 191 ].r := ScaleColorValue( GSwap^.R , t * 4 );
		MyPal[ T + 191 ].g := ScaleColorValue( GSwap^.G , t * 4 );
		MyPal[ T + 191 ].b := ScaleColorValue( GSwap^.B , t * 4 );
	end;
	SDL_SetPalette( MyImage , SDL_LOGPAL or SDL_PHYSPAL , MyPal , 0 , 256 );
end;

{$IFDEF PATCH_GH}
Function MakeSwapBitmapRYGA( MyImage: PSDL_Surface; RSwap,YSwap,GSwap: PSDL_Color ): PSDL_Surface;
	{ Swap those colors out for the requested colors. }
var
	MyImage2: PSDL_Surface;
	flags: UInt32;
	ret: Integer;
	bpp, w, h: Integer;
	pitch: Uint16;
	pixels: Pointer;
	x, y: Integer;
	p: PUint32;
	c: Uint32;
	rr, gg, bb, aa: Uint8;
	t: Integer;
begin
	{ Create replacement surface. }
	MyImage2 := SDL_CreateRGBSurface( ( SDL_SRCALPHA or SDL_SWSURFACE ) , MyImage^.W , MyImage^.H , 32 , $FF000000 , $00FF0000 , $0000FF00 , $000000FF );
	if ( NIL = MyImage2 ) then exit( NIL );

	{ Blit from the original to the copy. }
	flags := MyImage^.flags;
	SDL_SetAlpha( MyImage , ( MyImage^.flags and ( not SDL_SRCALPHA ) or SDL_RLEACCEL ) , SDL_ALPHA_OPAQUE );
	SDL_BlitSurface( MyImage , NIL , MyImage2 , NIL );
	SDL_SetAlpha( MyImage , flags , SDL_ALPHA_OPAQUE );
	SDL_SetAlpha( MyImage2 , ( MyImage2^.flags or SDL_RLEACCEL ) , SDL_ALPHA_OPAQUE );

	{ Swap colors }
	ret := SDL_LockSurface( MyImage2 );
	if ( 0 <> ret ) then begin
		ErrorMessage( 'Surface can not locked.' );
		exit( MyImage2 );
	end;
	bpp := MyImage2^.format^.BytesPerPixel;
	pixels := MyImage2^.pixels;
	pitch := MyImage2^.pitch;
	w := ( MyImage2^.W - 1 );
	h := ( MyImage2^.H - 1 );
	if ( 4 = bpp ) then begin
		for y := 0 to h do begin
			for x := 0 to w do begin
				p := pixels + y * pitch + x * bpp;
				c := p^;
				rr := ( ( c and $FF000000 ) shr 24 );
				gg := ( ( c and $00FF0000 ) shr 16 );
				bb := ( ( c and $0000FF00 ) shr  8 );
				aa := ( ( c and $000000FF ) shr  0 );

				if ( ( 0 = rr ) and ( 0 = gg ) and ( 0 = bb ) ) then begin
				end else if ( ( 0 = gg ) and ( 0 = bb ) ) then begin
					t := rr;
					rr := ScaleColorValue( RSwap^.R , t );
					gg := ScaleColorValue( RSwap^.G , t );
					bb := ScaleColorValue( RSwap^.B , t );
				end else if ( ( 0 = bb ) and ( rr = gg ) ) then begin
					t := rr;
					rr := ScaleColorValue( YSwap^.R , t );
					gg := ScaleColorValue( YSwap^.G , t );
					bb := ScaleColorValue( YSwap^.B , t );
				end else if ( ( 0 = rr ) and ( 0 = bb ) ) then begin
					t := gg;
					rr := ScaleColorValue( GSwap^.R , t );
					gg := ScaleColorValue( GSwap^.G , t );
					bb := ScaleColorValue( GSwap^.B , t );
				end;
				p^ := ( ( rr shl 24 ) or ( gg shl 16 ) or ( bb shl 8 ) or ( aa shl 0 ) );
			end;
		end;
	end else begin
		ErrorMessage( 'Surface is not 32bit RGB+alpha image.' );
	end;
	SDL_UnlockSurface( MyImage2 );

	MakeSwapBitmapRYGA := MyImage2;
end;

Function MakeSwapBitmapHSVA( MyImage: PSDL_Surface; RSwap,YSwap,GSwap: PSDL_Color ): PSDL_Surface;
	{ Swap those colors out for the requested colors. }
var
	MyImage2: PSDL_Surface;
	flags: UInt32;
	ret: Integer;
	bpp, w, h: Integer;
	pitch: Uint16;
	pixels: Pointer;
	x, y: Integer;
	p: PUint32;
	c: Uint32;
	rr, gg, bb, aa: Uint8;
	r, g, b: double;
	rgb_min: double;
	hsv_h, hsv_s, hsv_v: double;
	t: Integer;
begin
	{ Create replacement surface. }
	MyImage2 := SDL_CreateRGBSurface( ( SDL_SRCALPHA or SDL_SWSURFACE ) , MyImage^.W , MyImage^.H , 32 , $FF000000 , $00FF0000 , $0000FF00 , $000000FF );
	if ( NIL = MyImage2 ) then exit( NIL );

	{ Blit from the original to the copy. }
	flags := MyImage^.flags;
	SDL_SetAlpha( MyImage , ( MyImage^.flags and ( not SDL_SRCALPHA ) or SDL_RLEACCEL ) , SDL_ALPHA_OPAQUE );
	SDL_BlitSurface( MyImage , NIL , MyImage2 , NIL );
	SDL_SetAlpha( MyImage , flags , SDL_ALPHA_OPAQUE );
	SDL_SetAlpha( MyImage2 , ( MyImage2^.flags or SDL_RLEACCEL ) , SDL_ALPHA_OPAQUE );

	{ Swap colors }
	ret := SDL_LockSurface( MyImage2 );
	if ( 0 <> ret ) then begin
		ErrorMessage( 'Surface can not locked.' );
		exit( MyImage2 );
	end;
	bpp := MyImage2^.format^.BytesPerPixel;
	pixels := MyImage2^.pixels;
	pitch := MyImage2^.pitch;
	w := ( MyImage2^.W - 1 );
	h := ( MyImage2^.H - 1 );
	if ( 4 = bpp ) then begin
		for y := 0 to h do begin
			for x := 0 to w do begin
				p := pixels + y * pitch + x * bpp;
				c := p^;
				rr := ( ( c and $FF000000 ) shr 24 );
				gg := ( ( c and $00FF0000 ) shr 16 );
				bb := ( ( c and $0000FF00 ) shr  8 );
				aa := ( ( c and $000000FF ) shr  0 );

				r := ( float(rr) / 255.0 );
				g := ( float(gg) / 255.0 );
				b := ( float(bb) / 255.0 );
				hsv_v := Max( Max( r , g ) , b );
				rgb_min := Min( Min( r , g ) , b );
				hsv_h := ( hsv_v - rgb_min );
				hsv_s := hsv_h;
				if ( 0.0 < hsv_h ) then begin
					if ( hsv_v = r ) then begin
						hsv_h := ( g - b ) / hsv_h;
						if ( hsv_h < 0.0 ) then begin
							hsv_h := hsv_h + 6.0;
						end;
					end else if ( hsv_v = g ) then begin
						hsv_h := ( 2.0 + ( b - r ) / hsv_h );
					end else begin
						hsv_h := ( 4.0 + ( r - g ) / hsv_h );
					end;
				end;
				if ( 0.0 < hsv_v ) then begin
					hsv_s := ( hsv_s / hsv_v );
				end;

				t := trunc( hsv_v * 255.0 );
				if ( hsv_s < 0.5 ) then begin
					rr := t;
					gg := t;
					bb := t;
				end else if ( ( 0 = rr ) and ( 0 = gg ) ) then begin
					bb := 0;
					aa := 0;
				end else if ( ( 4.0 < hsv_h ) or ( hsv_h < 0.5 ) ) then begin
					rr := ScaleColorValue( RSwap^.R , t );
					gg := ScaleColorValue( RSwap^.G , t );
					bb := ScaleColorValue( RSwap^.B , t );
				end else if ( ( 0.5 <= hsv_h ) and ( hsv_h < 1.5 ) ) then begin
					rr := ScaleColorValue( YSwap^.R , t );
					gg := ScaleColorValue( YSwap^.G , t );
					bb := ScaleColorValue( YSwap^.B , t );
				end else begin
					rr := ScaleColorValue( GSwap^.R , t );
					gg := ScaleColorValue( GSwap^.G , t );
					bb := ScaleColorValue( GSwap^.B , t );
				end;
				p^ := ( ( rr shl 24 ) or ( gg shl 16 ) or ( bb shl 8 ) or ( aa shl 0 ) );
			end;
		end;
	end else begin
		ErrorMessage( 'Surface is not 32bit RGB+alpha image.' );
	end;
	SDL_UnlockSurface( MyImage2 );

	MakeSwapBitmapHSVA := MyImage2;
end;

Function MakeSwapBitmapBA( MyImage: PSDL_Surface ): PSDL_Surface;
	{ Swap those colors out for the requested colors. }
var
	MyImage2: PSDL_Surface;
	flags: UInt32;
	ret: Integer;
	bpp, w, h: Integer;
	pitch: Uint16;
	pixels: Pointer;
	x, y: Integer;
	p: PUint32;
	c: Uint32;
	rr, gg, bb, aa: Uint8;
begin
	{ Create replacement surface. }
	MyImage2 := SDL_CreateRGBSurface( ( SDL_SRCALPHA or SDL_SWSURFACE ) , MyImage^.W , MyImage^.H , 32 , $FF000000 , $00FF0000 , $0000FF00 , $000000FF );
	if ( NIL = MyImage2 ) then exit( NIL );

	{ Blit from the original to the copy. }
	flags := MyImage^.flags;
	SDL_SetAlpha( MyImage , ( MyImage^.flags and ( not SDL_SRCALPHA ) or SDL_RLEACCEL ) , SDL_ALPHA_OPAQUE );
	SDL_BlitSurface( MyImage , NIL , MyImage2 , NIL );
	SDL_SetAlpha( MyImage , flags , SDL_ALPHA_OPAQUE );
	SDL_SetAlpha( MyImage2 , ( MyImage2^.flags or SDL_RLEACCEL ) , SDL_ALPHA_OPAQUE );

	{ Swap colors }
	ret := SDL_LockSurface( MyImage2 );
	if ( 0 <> ret ) then begin
		ErrorMessage( 'Surface can not locked.' );
		exit( MyImage2 );
	end;
	bpp := MyImage2^.format^.BytesPerPixel;
	pixels := MyImage2^.pixels;
	pitch := MyImage2^.pitch;
	w := ( MyImage2^.W - 1 );
	h := ( MyImage2^.H - 1 );
	if ( 4 = bpp ) then begin
		for y := 0 to h do begin
			for x := 0 to w do begin
				p := pixels + y * pitch + x * bpp;
				c := p^;
				rr := ( ( c and $FF000000 ) shr 24 );
				gg := ( ( c and $00FF0000 ) shr 16 );
				bb := ( ( c and $0000FF00 ) shr  8 );
				aa := ( ( c and $000000FF ) shr  0 );

				if ( ( 0 = rr ) and ( 0 = gg ) ) then begin
					bb := 0;
					aa := 0;
				end else begin
				end;
				p^ := ( ( rr shl 24 ) or ( gg shl 16 ) or ( bb shl 8 ) or ( aa shl 0 ) );
			end;
		end;
	end else begin
		ErrorMessage( 'Surface is not 32bit RGB+alpha image.' );
	end;
	SDL_UnlockSurface( MyImage2 );

	MakeSwapBitmapBA := MyImage2;
end;
{$ENDIF PATCH_GH}

Procedure GenerateColor( var ColorString: String; var ColorStruct: TSDL_Color );
	{ Generate the color from the string. }
var
	n: Integer;
begin
	n := ExtractValue( ColorString );
	if n > 255 then n := 255;
	ColorStruct.R := n;
	n := ExtractValue( ColorString );
	if n > 255 then n := 255;
	ColorStruct.G := n;
	n := ExtractValue( ColorString );
	if n > 255 then n := 255;
	ColorStruct.B := n;
end;


Function LocateSpriteByNameColor( const name,color: String ): SensibleSpritePtr;
	{ Locate the sprite which matches the name provided. }
	{ If no such sprite exists, return Nil. }
var
	S: SensibleSpritePtr;
begin
	S := Game_Sprites;
	while ( S <> Nil ) and ( ( S^.Name <> name ) or ( S^.Color <> Color ) ) do begin
		S := S^.Next;
	end;
	LocateSpriteByNameColor := S;
end;

Function NewSprite: SensibleSpritePtr;
	{ Add an empty sprite description to the list. }
var
	it: SensibleSpritePtr;
begin
	New(it);
{$IFDEF DEBUG}
	Trace_MemoryLeak('NewSprite() New',it);
{$ENDIF DEBUG}
	if it = Nil then exit( Nil );
	{Initialize values.}
	it^.Next := Game_Sprites;
	it^.Color := '';
{$IFDEF PATCH_GH}
	it^.alpha_blending := False;
{$ENDIF PATCH_GH}
	Game_Sprites := it;
	NewSprite := it;
end;

Function AddSprite( name, color: String; W,H: Integer ): SensibleSpritePtr;
	{ Add a new element to the Sprite List. Load the image for this sprite }
	{ from disk, if possible. }
var
	fname: PChar;
	it: SensibleSpritePtr;
	tmp: PSDL_Surface;
	RSwap,YSwap,GSwap: TSDL_Color;
{$IFDEF PATCH_GH}
	org_name: String;
	ColorSwapMode: Boolean;
{$ENDIF PATCH_GH}
begin
	{Allocate memory for our new element.}
	it := NewSprite;
	if it = Nil then Exit( Nil );
	it^.Name := Name;
	it^.Color := Color;
	it^.W := W;
	it^.H := H;

{$IFDEF PATCH_GH}
	org_name := name;
{$ENDIF PATCH_GH}
	name := FSearch( name , Graphics_Directory );

	if name <> '' then begin
		fname := QuickPCopy( name );

		{ Attempt to load the image. }
		it^.Img := IMG_Load( fname );

		if it^.Img <> Nil then begin
{$IFDEF PATCH_GH}
			ColorSwapMode := False;
			if ( '' <> Color ) then begin
				ColorSwapMode := True;
				GenerateColor( Color , RSwap );
				GenerateColor( Color , YSwap );
				GenerateColor( Color , GSwap );
			end;
			if ( SDL_SRCALPHA = ( SDL_SRCALPHA and it^.Img^.flags ) ) then begin
				it^.alpha_blending := True;

				{ If a color swap has been specified, handle that here. }
				if ColorSwapMode then begin
					if ( 1 <= Pos( '.nsc.' , fname ) ) then begin
					end else if ( 1 <= Pos( '.psc.' , fname ) ) then begin
						tmp := MakeSwapBitmapRYGA( it^.Img , @RSwap , @YSwap , @GSwap );
						SDL_FreeSurface( it^.Img );
						it^.img := tmp;
					end else begin
						tmp := MakeSwapBitmapHSVA( it^.Img , @RSwap , @YSwap , @GSwap );
						SDL_FreeSurface( it^.Img );
						it^.img := tmp;
					end;
				end else begin
					if ( 1 <= Pos( '.nsc.' , fname ) ) then begin
					end else if ( 1 <= Pos( '.psc.' , fname ) ) then begin
						tmp := MakeSwapBitmapBA( it^.Img );
						SDL_FreeSurface( it^.Img );
						it^.img := tmp;
					end else begin
						tmp := MakeSwapBitmapBA( it^.Img );
						SDL_FreeSurface( it^.Img );
						it^.img := tmp;
					end;
				end;
			end else begin
				{ Set transparency color. }
				SDL_SetColorKey( it^.Img , SDL_SRCCOLORKEY or SDL_RLEACCEL , SDL_MapRGB( it^.Img^.Format , 0 , 0, 255 ) );

				{ If a color swap has been specified, handle that here. }
				if ColorSwapMode then begin
					if UseAdvancedColoring and ( it^.Img^.format^.palette <> Nil ) then begin
						RedefinePalette( it^.Img , @RSwap , @YSwap , @GSwap );
					end else begin
						tmp := MakeSwapBitmap( it^.Img , @RSwap , @YSwap , @GSwap );
						SDL_FreeSurface( it^.Img );
						it^.img := tmp;
					end;

				end;

				{ Convert to the screen mode. }
				{ This will make blitting far quicker. }
				tmp := SDL_ConvertSurface( it^.Img , Game_Screen^.Format , SDL_SRCCOLORKEY );
				SDL_FreeSurface( it^.Img );
				it^.Img := TMP;
			end;
{$ELSE PATCH_GH}
			{ Set transparency color. }
			SDL_SetColorKey( it^.Img , SDL_SRCCOLORKEY or SDL_RLEACCEL , SDL_MapRGB( it^.Img^.Format , 0 , 0, 255 ) );

			{ If a color swap has been specified, handle that here. }
			if Color <> '' then begin
				GenerateColor( Color , RSwap );
				GenerateColor( Color , YSwap );
				GenerateColor( Color , GSwap );

				if UseAdvancedColoring and ( it^.Img^.format^.palette <> Nil ) then begin
					RedefinePalette( it^.Img , @RSwap , @YSwap , @GSwap );
				end else begin
					tmp := MakeSwapBitmap( it^.Img , @RSwap , @YSwap , @GSwap );
					SDL_FreeSurface( it^.Img );
					it^.img := tmp;
				end;

			end;

			{ Convert to the screen mode. }
			{ This will make blitting far quicker. }
			tmp := SDL_ConvertSurface( it^.Img , Game_Screen^.Format , SDL_SRCCOLORKEY );
			SDL_FreeSurface( it^.Img );
			it^.Img := TMP;
{$ENDIF PATCH_GH}
		end;

{$IF DEFINED(PATCH_I18N) or DEFINED(PATCH_GH) or DEFINED(WITH_TENC)}
		sysutils.StrDispose( fname );
{$ELSE}
		Dispose( fname );
{$ENDIF}
	end else begin
		it^.Img := Nil;

	end;

{$IFDEF PATCH_GH}
	if ( '' <> org_name ) and ( NIL = it^.Img ) then begin
		ErrorMessage_fork('Sprite data "' + org_name + '" is not found.');
	end;
{$ENDIF PATCH_GH}

	{Return a pointer to the new element.}
	AddSprite := it;
end;

Procedure DisposeSpriteList(var LList: SensibleSpritePtr);
	{Dispose of the list, freeing all associated system resources.}
var
	LTemp: SensibleSpritePtr;
begin
	while LList <> Nil do begin
		LTemp := LList^.Next;

{$IFDEF DEBUG}
		if LList^.Img <> Nil then begin
			Trace_MemoryLeak('DisposeSpriteList() Dispose Img',LList^.Img);
			CheckAndNIL_Pointer('DisposeSpriteList() Dispose',LList^.Img,True);
			SDL_FreeSurface( LList^.Img );
			LList^.Img := NIL;
		end;
{$ELSE DEBUG}
		if LList^.Img <> Nil then SDL_FreeSurface( LList^.Img );
{$ENDIF DEBUG}

{$IFDEF DEBUG}
		Trace_MemoryLeak('DisposeSpriteList() Dispose',LList);
		CheckAndNIL_Pointer('DisposeSpriteList() Dispose',LList,True);
{$ENDIF DEBUG}
{$IFDEF PATCH_GH_PARANOID_SAFER}
		LList^.Name[1]  := '@';
		LList^.Color[1] := '@';
		LList^.W        := -32767;
		LList^.H        := -32767;
{$IFDEF PATCH_GH}
		LList^.alpha_blending := False;
{$ENDIF PATCH_GH}
		LList^.Img      := PSDL_Surface(-1);
		LList^.Next     := SensibleSpritePtr(-1);
{$ENDIF PATCH_GH_PARANOID_SAFER}
		Dispose(LList);
		LList := LTemp;
	end;
end;


Procedure RemoveSprite(var LMember: SensibleSpritePtr);
	{Locate and extract member LMember from list LList.}
	{Then, dispose of LMember.}
var
	a,b: SensibleSpritePtr;
begin
	{Initialize A and B}
	B := Game_Sprites;
	A := Nil;

	{Locate LMember in the list. A will thereafter be either Nil,}
	{if LMember if first in the list, or it will be equal to the}
	{element directly preceding LMember.}
	while (B <> LMember) and (B <> Nil) do begin
		A := B;
		B := B^.next;
	end;

	if B = Nil then begin
		{Major FUBAR. The member we were trying to remove can't}
		{be found in the list.}
{$IFDEF PATCH_GH}
		ErrorMessage_fork('ERROR- RemoveLink asked to remove a link that doesnt exist.');
{$ELSE PATCH_GH}
		writeln('ERROR- RemoveLink asked to remove a link that doesnt exist.');
{$ENDIF PATCH_GH}
		end
	else if A = Nil then begin
		{There's no element before the one we want to remove,}
		{i.e. it's the first one in the list.}
		Game_Sprites := B^.Next;
		B^.Next := Nil;
		DisposeSpriteList(B);
		end
	else begin
		{We found the attribute we want to delete and have another}
		{one standing before it in line. Go to work.}
		A^.next := B^.next;
		B^.Next := Nil;
		DisposeSpriteList(B);
	end;
end;

Procedure CleanSpriteList;
	{ Go through the sprite list and remove those sprites we aren't likely to }
	{ need immediately... i.e., erase those ones which have a COLOR string defined. }
var
	S,S2: SensibleSpritePtr;
begin
	S := Game_Sprites;
	while S <> Nil do begin
		S2 := S^.Next;

		if ( S^.Color <> '' ) or ( S^.img = Nil ) then begin
			RemoveSprite( S );
		end;

		S := S2;
	end;
end;

Procedure DrawAnimImage( Image: PSDL_Surface; W,H,Frame: Integer; var MyDest: TSDL_Rect );
	{ This procedure is modeled after the command from Blitz Basic. }
var
	MySource: TSDL_Rect;
begin
	MySource.W := W;
	MySource.H := H;
	if W > Image^.W then W := Image^.W;
	MySource.X := ( Frame mod ( Image^.W div W ) ) * W;
	MySource.Y := ( Frame div ( Image^.W div W ) ) * H;

	SDL_BlitSurface( Image , @MySource , Game_Screen , @MyDest );
end;

{$IFDEF PATCH_GH}
Procedure DrawAnimImageRGBA( Image: PSDL_Surface; W,H,Frame: Integer; var MyDest: TSDL_Rect );
	{ This procedure is modeled after the command from Blitz Basic. }
var
	MySource: TSDL_Rect;
begin
	MySource.W := W;
	MySource.H := H;
	if W > Image^.W then W := Image^.W;
	MySource.X := ( Frame mod ( Image^.W div W ) ) * W;
	MySource.Y := ( Frame div ( Image^.W div W ) ) * H + ( Image^.H div 2 );

	SDL_BlitSurface( Image , @MySource , Game_Screen , @MyDest );
end;
{$ENDIF PATCH_GH}

procedure DrawSprite( Spr: SensibleSpritePtr; MyDest: TSDL_Rect; Frame: Integer );
	{ Draw a sensible sprite. }
begin
	{ First make sure that we have some valid sprite data... }
	if ( Spr <> Nil ) and ( Spr^.Img <> Nil ) then begin
		{ All the info checks out. Print it. }
		DrawAnimImage( Spr^.Img , Spr^.W , Spr^.H , Frame , MyDest );
	end;
end;

procedure DrawAlphaSprite( Spr: SensibleSpritePtr; MyDest: TSDL_Rect; Frame: Integer );
	{ Draw a sensible sprite. }
begin
	{ First make sure that we have some valid sprite data... }
	if ( Spr <> Nil ) and ( Spr^.Img <> Nil ) then begin
{$IFDEF PATCH_GH}
		if ( Spr^.alpha_blending ) then begin
			DrawAnimImageRGBA( Spr^.Img , Spr^.W , Spr^.H , Frame , MyDest );
		end else begin
{$ENDIF PATCH_GH}
		{ All the info checks out. Print it. }
		SDL_SetAlpha( Spr^.Img , SDL_SRCAlpha , Alpha_Level );
		DrawAnimImage( Spr^.Img , Spr^.W , Spr^.H , Frame , MyDest );
		SDL_SetAlpha( Spr^.Img , SDL_SRCAlpha , SDL_Alpha_Opaque );
{$IFDEF PATCH_GH}
		end;
{$ENDIF PATCH_GH}
	end;
end;

Function ConfirmSprite( Name: String; const Color: String; W,H: Integer ): SensibleSpritePtr;
	{ Try to locate the requested sprite in the requested color. If the sprite }
	{ is already loaded, then return its address. If not, load it and color it. }
var
	S: SensibleSpritePtr;
begin
	{ First, find the sprite. If by some strange chance it hasn't been }
	{ loaded yet, load it now. }
	S := LocateSpriteByNameColor( Name , Color );
	if S = Nil then S := AddSprite( Name , Color , W , H );

	{ Set the width and height fields. }
	S^.W := W;
	S^.H := H;

	ConfirmSprite := S;
end;


{$IFDEF PATCH_AUDIO}
Procedure Load_SoundEffect;
var
	SoundEffectFileList: SAttPtr;
	P: SAttPtr;
	S: SensibleSoundEffectListPtr;
	PText: PChar;
begin
	SoundEffectFileList := CreateFileList( Sound_Directory + '*.*' );
	SoundEffectList := NIL;

	P := SoundEffectFileList;
	while ( NIL <> P ) do begin
		if ( ( 3 <= Length( P^.Info ) ) and ( '.txt' <> Copy( P^.Info , (Length(P^.Info) - 3) , 4 ) ) and ( '.bak' <> Copy( P^.Info , (Length(P^.Info) - 3) , 4 ) ) ) then begin
			New( S );
{$IFDEF DEBUG}
			Trace_MemoryLeak('Load_SoundEffect() New',S);
{$ENDIF DEBUG}
			if ( NIL = S ) then exit;

			S^.Name := P^.Info;
			PText := QuickPCopy( Sound_Directory + P^.Info );
			S^.Chunk := Mix_LoadWAV( PText );
{$IF DEFINED(PATCH_I18N) or DEFINED(PATCH_GH) or DEFINED(WITH_TENC)}
			sysutils.StrDispose( PText );
{$ELSE}
			Dispose( PText );
{$ENDIF}
			S^.Next := SoundEffectList;

			if ( NIL <> S^.Chunk ) then begin
				SoundEffectList := S;
			end else begin
				ErrorMessage( 'SDL_mixer error at Mix_LoadWAV(' + P^.Info + ') : ' + SDL_GetError );
{$IFDEF DEBUG}
				Trace_MemoryLeak('Load_SoundEffect() Dispose',S);
				CheckAndNIL_Pointer('Load_SoundEffect() Dispose',S,True);
{$ENDIF DEBUG}
{$IFDEF PATCH_GH_PARANOID_SAFER}
				S^.Name[1] := '@';
				S^.Chunk   := NIL;
				S^.Next    := SensibleSoundEffectListPtr(-1);
{$ENDIF PATCH_GH_PARANOID_SAFER}
				Dispose( S );
			end;
		end;
		P := P^.Next;
	end;

	DisposeSAtt( SoundEffectFileList );
end;

Procedure Dispose_SoundEffect;
var
	S,S2: SensibleSoundEffectListPtr;
begin
	S := SoundEffectList;
	while ( NIL <> S ) do begin
		S2 := S^.Next;

		Mix_FreeChunk( S^.Chunk );
{$IFDEF DEBUG}
		Trace_MemoryLeak('Dispose_SoundEffect() Dispose',S);
		CheckAndNIL_Pointer('Dispose_SoundEffect() Dispose',S,True);
{$ENDIF DEBUG}
{$IFDEF PATCH_GH_PARANOID_SAFER}
		S^.Name[1] := '@';
		S^.Chunk   := NIL;
		S^.Next    := SensibleSoundEffectListPtr(-1);
{$ENDIF PATCH_GH_PARANOID_SAFER}
		Dispose( S );

		S := S2;
	end;
end;

Procedure PlaySoundEffect_Direct( const Name: String );
var
	S: SensibleSoundEffectListPtr;
begin
	if ( Length(Name) < 3 ) then exit;

	S := SoundEffectList;
	while ( NIL <> S ) do begin
		if ( Name = S^.Name ) then begin
			Mix_PlayChannel( -1 , S^.Chunk , 0 );
			exit;
		end;
		S := S^.Next;
	end;
	ErrorMessage_fork( 'A sound effect file "' + Name + '" was not found.' );
end;
{$ENDIF PATCH_AUDIO}


{$IFDEF PATCH_I18N}
function RPGKey( var Unicode: Word ): Char;
begin
	RPGKey := RPGKey( Unicode, True );
end;

function RPGKey: Char;
var
	dummy: Word;
begin
	RPGKey := RPGKey( dummy, False );
end;
{$ENDIF PATCH_I18N}

{$IFDEF PATCH_I18N}
function RPGKey( var Unicode: Word; const GetStringMode: Boolean ): Char;
{$ELSE PATCH_I18N}
function RPGKey: Char;
{$ENDIF PATCH_I18N}
	{ Read a readable key from the keyboard and return its ASCII value. }
var
	a: String;
	event : TSDL_Event;
	m2: PChar;
	width,height: Integer;
	pmsg: PChar;
{$IFDEF PATCH_GH}
	t: LongInt;
	MouseMoved: Boolean = False;
	i: Integer;
{$ENDIF PATCH_GH}
begin
	a := '';
{$IFDEF PATCH_I18N}
	Unicode := 0;
{$ENDIF PATCH_I18N}
	repeat
		{ Wait for events. }
		if SDL_PollEvent( @event ) = 1 then begin
			{ See if this event is a keyboard one... }
			if event.type_ = SDL_KEYDOWN then begin
				{ Check to see if it was an ASCII character we received. }
{$IF DEFINED(PATCH_GH) and DEFINED(PATCH_I18N)}
				if GetStringMode then begin
					if ( 0 < event.key.keysym.unicode ) and ( event.key.keysym.unicode < $80 ) then begin
						a := Char( event.key.keysym.unicode );
					end else begin
						Unicode := event.key.keysym.unicode;
						a := #$80;
					end;
				end else begin
					case event.key.keysym.sym of
						SDLK_Backspace:	a := KBD_BS;
						SDLK_Tab:	a := KBD_TAB;
						SDLK_Return:	a := KBD_RET;
						SDLK_Escape:	a := KBD_ESC;
						SDLK_Delete:	a := KBD_Delete;
						SDLK_KP0:		a := KBD_Insert;
						SDLK_KP1:		a := KBD_End;
						SDLK_KP2:		a := KBD_Down;
						SDLK_KP3:		a := KBD_PgDn;
						SDLK_KP4:		a := KBD_Left;
						SDLK_KP5:		a := KBD_KP5;
						SDLK_KP6:		a := KBD_Right;
						SDLK_KP7:		a := KBD_Home;
						SDLK_KP8:		a := KBD_Up;
						SDLK_KP9:		a := KBD_PgUp;
						SDLK_KP_Period:		a := KBD_Delete;
						SDLK_KP_Divide:		a := '/';
						SDLK_KP_Multiply:	a := '*';
						SDLK_KP_Minus:		a := '-';
						SDLK_KP_Plus:		a := '+';
						SDLK_KP_Enter:		a := KBD_RET;
						SDLK_KP_Equals:		a := '=';
						SDLK_Up:	a := KBD_Up;
						SDLK_Down:	a := KBD_Down;
						SDLK_Right:	a := KBD_Right;
						SDLK_Left:	a := KBD_Left;
						SDLK_Insert:	a := KBD_Insert;
						SDLK_Home:	a := KBD_Home;
						SDLK_End:	a := KBD_End;
						SDLK_PageUp:	a := KBD_PgUp;
						SDLK_PageDown:	a := KBD_PgDn;
						SDLK_F1:	a := KBD_F1;
						SDLK_F2:	a := KBD_F2;
						SDLK_F3:	a := KBD_F3;
						SDLK_F4:	a := KBD_F4;
						SDLK_F5:	a := KBD_F5;
						SDLK_F6:	a := KBD_F6;
						SDLK_F7:	a := KBD_F7;
						SDLK_F8:	a := KBD_F8;
						SDLK_F9:	a := KBD_F9;
						SDLK_F10:	a := KBD_F10;
						SDLK_F11:	a := KBD_F11;
						SDLK_F12:	a := KBD_F12;
						SDLK_F13:	a := KBD_F13;
						SDLK_F14:	a := KBD_F14;
						SDLK_F15:	a := KBD_F15;
					else
						if ( 0 < event.key.keysym.unicode ) and ( event.key.keysym.unicode < $80 ) then begin
							a := Char( event.key.keysym.unicode );
						end;
					end;
				end;
{$ELSE}
				case event.key.keysym.sym of
				SDLK_F1:	begin
							pmsg := QuickPCopy( Config_Directory + replacehash('Demo#.bmp',Bstr(animation_phase)) );
							SDL_SaveBmp( Game_Screen , pmsg );
  {$IF DEFINED(PATCH_I18N) or DEFINED(PATCH_GH) or DEFINED(WITH_TENC)}
							sysutils.StrDispose( pmsg );
  {$ELSE}
							Dispose( pmsg );
  {$ENDIF}
						end;
  {$IFDEF PATCH_JPSSDL}
					SDLK_F9:		if SkipAnim then SkipAnim := False else SkipAnim := True;
  {$ENDIF PATCH_JPSSDL}
  {$IFDEF PATCH_CHEAT}
					SDLK_F12:		if Cheat_Display then if Cheat_Display_SW then Cheat_Display_SW := False else Cheat_Display_SW := True;
  {$ENDIF PATCH_CHEAT}
					SDLK_Up,SDLK_KP8:	a := RPK_Up;
					SDLK_Down,SDLK_KP2:	a := RPK_Down;
					SDLK_Left,SDLK_KP4:	a := RPK_Left;
					SDLK_Right,SDLK_KP6:	a := RPK_Right;
					SDLK_KP7:		a := RPK_UpLeft;
					SDLK_KP9:		a := RPK_UpRight;
					SDLK_KP1:		a := RPK_DownLeft;
					SDLK_KP3:		a := RPK_DownRight;
					SDLK_Backspace:		a := #8;
					SDLK_KP_Enter:		a := #10;
					SDLK_KP5:		a := '5';
				else
  {$IFDEF PATCH_I18N}
					if( event.key.keysym.unicode <  $80 ) and ( event.key.keysym.unicode > 0 ) then begin
						a := Char( event.key.keysym.unicode );
					end else if (True = GetStringMode) and (event.key.keysym.unicode > 0) then begin
						Unicode := event.key.keysym.unicode;
						a := #$80;
					end;
  {$ELSE PATCH_I18N}
					if( event.key.keysym.unicode <  $80 ) and ( event.key.keysym.unicode > 0 ) then begin
						a := Char( event.key.keysym.unicode );
					end;
  {$ENDIF PATCH_I18N}
				end;
{$ENDIF}

			end else if ( event.type_ = SDL_MOUSEButtonDown ) then begin
				{ Return a mousebutton event. }
{$IFDEF PATCH_GH}
				case event.button.button of
				SDL_BUTTON_LEFT:		a := RPK_MouseButton;
				SDL_BUTTON_MIDDLE:		a := KeyMap[ KMC_ButtonMiddle ].KCode;
				SDL_BUTTON_RIGHT:		a := RPK_RightButton;
				SDL_BUTTON_WHEELUP:		a := KeyMap[ KMC_ButtonWUp ].KCode;
				SDL_BUTTON_WHEELDOWN:		a := KeyMap[ KMC_ButtonWDown ].KCode;
				(SDL_BUTTON_WHEELDOWN+1):	a := KeyMap[ KMC_ButtonWLeft ].KCode; { SDL_BUTTON_X1: (WHEEL-LEFT) }
				(SDL_BUTTON_WHEELDOWN+2):	a := KeyMap[ KMC_ButtonWRight ].KCode; { SDL_BUTTON_X2: (WHEEL-RIGHT) }
				end;
{$ELSE PATCH_GH}
				if event.button.button = SDL_BUTTON_LEFT then begin
					a := RPK_MouseButton;
				end else if event.button.button = SDL_BUTTON_RIGHT then begin
					a := RPK_RightButton;
				end;
{$ENDIF PATCH_GH}

{$IFDEF PATCH_GH}
			end else if ( SDL_MOUSEButtonUp = event.type_ ) then begin
				case event.button.button of
				SDL_BUTTON_LEFT:		a := RPK_MouseButtonRelease;
				end;

			end else if ( SDL_MOUSEMOTION = event.type_ ) then begin
				MouseMoved := True;
{$ENDIF PATCH_GH}

			end else if event.type_ = SDL_VIDEORESIZE then begin
				width := event.resize.w;
				if width < 800 then width := 800;
				height := event.resize.h;
				if height < 600 then height := 600;
{$IFDEF PATCH_GH}
				ScreenSize_Width  := width;
				ScreenSize_Height := height;
				Recalc_DynamicRect;
{$ENDIF PATCH_GH}
{$IFDEF PATCH_GH}
				Game_Screen := SDL_SetVideoMode(width, height, 0, SDL_VideoModeFlag );
{$ELSE PATCH_GH}
				Game_Screen := SDL_SetVideoMode(width, height, 0, SDL_HWSURFACE or SDL_DoubleBuf or SDL_RESIZABLE );
{$ENDIF PATCH_GH}
{$IFDEF PATCH_GH}
				MouseMoved := True;
{$ENDIF PATCH_GH}

			end;

		end else begin
{$IFDEF PATCH_GH}
			if 0 = Last_Clock_Update then begin
				Last_Clock_Update := SDL_GetTicks - FrameInterval;
			end;
			if SDL_GetTicks < ( Last_Clock_Update + FrameInterval - SleepGranularity ) then begin
				t := Last_Clock_Update + FrameInterval - SDL_GetTicks;
				if 0 < t then begin
					if t > KeysamplingInterval then SDL_Delay( KeysamplingInterval )
					else SDL_Delay( t );
				end;
			end;
			while (SDL_GetTicks >= (Last_Clock_Update + FrameInterval - 1)) and (FrameSkip or FrameDrawDone) do begin
				Last_Clock_Update := Last_Clock_Update + FrameInterval;
				Animation_Phase := ( Animation_Phase + 1 ) mod 6000;
				FrameDrawDone := False;
				a := RPK_TimeEvent;
			end;
{$ELSE PATCH_GH}
			if SDL_GetTicks < ( Last_Clock_Update + 20 ) then SDL_Delay( Last_Clock_Update + 30 - SDL_GetTicks );
			Last_Clock_Update := SDL_GetTicks + 30;
			Animation_Phase := ( Animation_Phase + 1 ) mod 6000;
			a := RPK_TimeEvent;
{$ENDIF PATCH_GH}

		end;

{$IFDEF PATCH_GH}
		if ('' <> a) then begin
			if RPGKey_LastEvent <> a then begin
				RPGKey_LastEvent := a;
			end else if ( RPK_TimeEvent = a ) then begin
				if SDL_NoTimeEvent then begin
					if MouseMoved then begin
						MouseMoved := False;
						a := RPK_MouseMotion;
					end else begin
						a := '';
					end;
				end else begin
					if MouseMoved then begin
						MouseMoved := False;
						a := RPK_MouseMotion;
					end;
				end;
			end;
		end;
{$ENDIF PATCH_GH}

	{ Keep going until either a character is found, or an error is reported. }
	until ( a <> '' );

	{ Possibly load music now. }
{	if ( Music_List <> Nil ) and ( Mix_PlayingMusic() = 0 ) then begin
		if MyMusic <> Nil then MIX_FreeMusic( MyMusic );
		m2 := QuickPCopy( SelectRandomSAtt( Music_List )^.info );
		MyMusic := MIX_LoadMus( m2 );
		Dispose( m2 );
		Mix_PlayMusic( MyMusic , 1 );
	end;}

{$IFDEF PATCH_GH}
	if not GetStringMode then begin
		i := 1;
		while ( KBD_ERROR <> KCodeAlias[i].KCode_from ) do begin
			if ( KCodeAlias[i].KCode_from = a ) then begin
				a := KCodeAlias[i].KCode_to;
				break;
			end;
			Inc( i );
		end;
	end;
{$ENDIF PATCH_GH}

	if a <> '' then RPGKey := a[1]
{$IFDEF PATCH_GH}
	else RPGKey := KBD_NONE;
{$ELSE PATCH_GH}
	else RPGKey := 'Z';
{$ENDIF PATCH_GH}
end;

Procedure ClrZone( var Z: TSDL_Rect );
	{ Clear the specified screen zone. }
begin
	SDL_FillRect( game_screen , @Z , SDL_MapRGB( Game_Screen^.Format , 0 , 0 , 0 ) );
end;

Procedure ClrScreen;
	{ Clear the specified screen zone. }
begin
	SDL_FillRect( game_screen , Nil , SDL_MapRGB( Game_Screen^.Format , 0 , 0 , 0 ) );
end;

Function TextLength( F: PTTF_Font; const msg: String ): LongInt;
	{ Determine how long "msg" will be using the default "game_font". }
var
{$IFDEF PATCH_I18N}
	pWC: PUInt16;
{$ELSE PATCH_I18N}
	pmsg: PChar;	{ Gotta convert to pchar, pain in the ass... }
{$ENDIF PATCH_I18N}
	W,Y: LongInt;	{ W means width I guess... Y is anyone's guess. Height? }
begin
{$IFDEF PATCH_I18N}
	pWC := New_Conv_ToUni16( msg );
	TTF_SizeUnicode( F , pWC , W , Y );
	sysutils.StrDispose( PChar( pWC ) );
{$ELSE PATCH_I18N}
	{ Convert the string to a pchar. }
	pmsg := QuickPCopy( msg );

	{ Call the alleged size calculation function. }
	TTF_SizeText( F , pmsg , W , Y );

	{ get rid of the PChar, since it's served its usefulness. }
  {$IF DEFINED(PATCH_I18N) or DEFINED(PATCH_GH) or DEFINED(WITH_TENC)}
	sysutils.StrDispose( pmsg );
  {$ELSE}
	Dispose( pmsg );
  {$ENDIF}
{$ENDIF PATCH_I18N}

	TextLength := W;
end;

{$IFDEF PATCH_I18N}
Procedure GetNextLine( var TheLine , msg , NextWord: String; Width: Integer; MyFont: PTTF_Font );
	{ Get a line of text of maximum width "Width". }
var
	LC: Boolean;	{ Loop Condition. So I wasn't very creative when I named it, so what? }
	BW: String;
	CW_I18N: Boolean;	{Is the current word I18N ?}
	DItS: Boolean;		{Do insert the space, or not.}
begin
	{ Loop condition starts out as TRUE. }
	LC := True;
	if TheLine = ' ' then TheLine := '';

	{ Start building the line. }
	repeat
		NextWord := ExtractWordForPrint( Msg, DItS, CW_I18N );

		if '' <> NextWord then begin
			if False = CW_I18N then begin
				if TextLength( MyFont , TheLine + ' ' + NextWord + I18N_Width_Of_One_Character + ' ') < Width then begin
					if DItS then TheLine := TheLine + ' ' + NextWord
					else TheLine := TheLine + NextWord
				end else begin
					LC := False;
				end;
			end else begin
				if TextLength( MyFont , TheLine + ' ' + NextWord + I18N_Width_Of_One_Character + ' ') < Width then begin
					if DItS then TheLine := TheLine + ' ' + NextWord
					else TheLine := TheLine + NextWord
				end else begin
					LC := False;

					if Pos(NextWord, ProhibitationHead) > 0 then begin
						TheLine := TheLine + NextWord + #13;
					end else begin
						BW := TailMBChar(TheLine);
						if (0 < Length(BW)) and (0 < Pos(BW, ProhibitationTrail)) then begin
							TheLine := Copy(TheLine,1,Length(TheLine)-Length(BW));
							NextWord := BW + NextWord;
						end;
					end;
				end;
			end;
		end;
	until (not LC) or (NextWord = '') or ( (0 < Length(TheLine)) and (TheLine[Length(TheLine)] = #13) );

	{ If the line ended due to a line break, deal with it. }
	if ( TheLine[Length(TheLine)] = #13 ) then begin
		{ Display the line break as a space. }
		TheLine[Length(TheLine)] := ' ';
		NextWord := ExtractWordForPrint( Msg, DItS, CW_I18N );
	end;

end;
{$ELSE PATCH_I18N}
Procedure GetNextLine( var TheLine , msg , NextWord: String; Width: Integer; MyFont: PTTF_Font );
	{ Get a line of text of maximum width "Width". }
var
	LC: Boolean;	{ Loop Condition. So I wasn't very creative when I named it, so what? }
begin
	{ Loop condition starts out as TRUE. }
	LC := True;

	{ Start building the line. }
	repeat
		NextWord := ExtractWord( Msg );

		if TextLength( MyFont , THEline + ' ' + NextWord) < Width then
			THEline := THEline + ' ' + NextWord
		else
			LC := False;

	until (not LC) or (NextWord = '') or ( TheLine[Length(TheLine)] = #13 );

	{ If the line ended due to a line break, deal with it. }
	if ( TheLine[Length(TheLine)] = #13 ) then begin
		{ Display the line break as a space. }
		TheLine[Length(TheLine)] := ' ';
		NextWord := ExtractWord( msg );
	end;

end;
{$ENDIF PATCH_I18N}

{$IFDEF PATCH_I18N}
Function New_Conv_ToUni16( const arg_msg: String ): PUInt16;
const
	WCLen = 576; { 512; }
var
	pmsg: PChar;
	pdst: PWord;
begin
	pmsg := QuickPCopy(arg_msg);
	pdst := PWord(sysutils.StrAlloc(WCLen));
	Conv_ToUni16( pmsg, Length(arg_msg), pdst, WCLen );
	New_Conv_ToUni16 := pdst;
	sysutils.StrDispose( pmsg );
end;

Function I18N_TTF_RenderText(var font: PTTF_Font; const pline: String; var fg: TSDL_Color ): PSDL_Surface;
var
	pWC: PUInt16;
	bg: TSDL_Color;
begin
	if TERMINAL_bidiRTL then begin
		pWC := New_Conv_ToUni16( Conv_bidiRTL(pline) );
	end else begin
		pWC := New_Conv_ToUni16( pline );
	end;
	if not(SDL_AAFont) and not(SDL_AAFont_Shaded) then begin
		I18N_TTF_RenderText := TTF_RenderUnicode_Solid( font, pWC, fg );
	end else if SDL_AAFont_Shaded then begin
		bg.r := 0;
		bg.b := 0;
		bg.g := 0;
		I18N_TTF_RenderText := TTF_RenderUnicode_Shaded( font, pWC, fg, bg );
	end else begin
		I18N_TTF_RenderText := TTF_RenderUnicode_Blended( font, pWC, fg );
	end;
	sysutils.StrDispose( PChar( pWC ) );
end;
{$ENDIF PATCH_I18N}

{Can't const}
Function PrettyPrint( msg: string; Width: Integer; var FG: TSDL_Color; DoCenter: Boolean; MyFont: PTTF_Font ): PSDL_Surface;
	{ Create a SDL_Surface containing all the text within "msg" formatted }
	{ in lines of no longer than "width" pixels. Sound simple? Mostly just }
	{ tedious, I'm afraid. }
var
	SList,SA: SAttPtr;
	S_Total,S_Temp: PSDL_Surface;
	MyDest: SDL_Rect;
{$IFDEF PATCH_I18N}
	rmask, gmask, bmask, amask: UInt32;
{$ELSE PATCH_I18N}
	pline: PChar;
{$ENDIF PATCH_I18N}
	NextWord: String;
	THELine: String;	{The line under construction.}
begin
	{ CLean up the message a bit. }
	DeleteWhiteSpace( msg );
	if msg = '' then Exit( Nil );

	{THELine = The first word in this iteration}
{$IFDEF PATCH_I18N}
	TheLine := ' ';
{$ELSE PATCH_I18N}
	THELine := ExtractWord( msg );
{$ENDIF PATCH_I18N}
	NextWord := '';
	SList := Nil;

	{Start the main processing loop.}
	while TheLine <> '' do begin
		GetNextLine( TheLine , msg , NextWord , Width, MyFont );

		{ Output the line. }
		{ Next append it to whatever has already been created. }
		StoreSAtt( SList , TheLine );

		{ Prepare for the next iteration. }
		TheLine := NextWord;
	end; { while TheLine <> '' }

	{ Create a bitmap for the message. }
	if SList <> Nil then begin
		{ Create a big bitmap to hold everything. }
{$IFDEF PATCH_I18N}
		rmask := $00FF0000;
		gmask := $0000FF00;
		bmask := $000000FF;
		amask := $FF000000;
		if not(SDL_AAFont) and not(SDL_AAFont_Shaded) then begin
			S_Total := SDL_CreateRGBSurface( SDL_SWSURFACE , width , TTF_FontLineSkip( MyFont ) * NumSAtts( SList ) , 32 , rmask, gmask, bmask, amask );
		end else if SDL_AAFont_Shaded then begin
			S_Total := SDL_CreateRGBSurface( SDL_SWSURFACE , width , TTF_FontLineSkip( MyFont ) * NumSAtts( SList ) , 32 , rmask, gmask, bmask, amask );
		end else begin
			S_Total := SDL_CreateRGBSurface( ( SDL_SRCALPHA or SDL_SWSURFACE ) , width , TTF_FontLineSkip( MyFont ) * NumSAtts( SList ) , 32 , rmask, gmask, bmask, amask );
		end;
{$ELSE PATCH_I18N}
{		S_Total := SDL_CreateRGBSurface( SDL_SWSURFACE , width , TTF_FontLineSkip( MyFont ) * NumSAtts( SList ) , 16 , 0 , 0 , 0 , 0 );
}		S_Total := SDL_CreateRGBSurface( SDL_SWSURFACE , width , TTF_FontLineSkip( MyFont ) * NumSAtts( SList ) , 32 , $FF000000 , $00FF0000 , $0000FF00 , $000000FF );
{$ENDIF PATCH_I18N}
		MyDest.X := 0;
		MyDest.Y := 0;

		{ Add each stored string to the bitmap. }
		SA := SList;
		while SA <> Nil do begin
{$IFDEF PATCH_I18N}
			S_Temp := I18N_TTF_RenderText( MyFont , SA^.Info , fg );
			if not(SDL_AAFont) and not(SDL_AAFont_Shaded) then begin
			end else if SDL_AAFont_Shaded then begin
			end else begin
				SDL_SetAlpha( S_Temp , 0 , SDL_ALPHA_OPAQUE );
			end;
{$ELSE PATCH_I18N}
			pline := QuickPCopy( SA^.Info );
			S_Temp := TTF_RenderText_Solid( MyFont , pline , fg );
{$IFDEF Unix}
			SDL_SetColorKey( S_Temp , SDL_SRCCOLORKEY , SDL_MapRGB( S_Temp^.Format , 0 , 0, 0 ) );
{$ENDIF}

  {$IF DEFINED(PATCH_I18N) or DEFINED(PATCH_GH) or DEFINED(WITH_TENC)}
			sysutils.StrDispose( pline );
  {$ELSE}
			Dispose( pline );
  {$ENDIF}
{$ENDIF PATCH_I18N}

			{ We may or may not be required to do centering of the text. }
			if DoCenter then begin
				MyDest.X := ( Width - TextLength( MyFont , SA^.Info ) ) div 2;
			end else begin
{$IFDEF PATCH_I18N}
				MyDest.X := 0;
				if TERMINAL_bidiRTL then begin
					MyDest.X := Width - S_Temp^.W;
				end;
{$ELSE PATCH_I18N}
				MyDest.X := 0;
{$ENDIF PATCH_I18N}
			end;

			SDL_BlitSurface( S_Temp , Nil , S_Total , @MyDest );
			SDL_FreeSurface( S_Temp );
			MyDest.Y := MyDest.Y + TTF_FontLineSkip( MyFont );
			SA := SA^.Next;
		end;
		DisposeSAtt( SList );

	end else begin
		S_Total := Nil;
	end;


	PrettyPrint := S_Total;
end;

Procedure QuickText( const msg: String; MyDest: TSDL_Rect; Color: TSDL_Color );
	{ Quickly draw some text to the screen, without worrying about }
	{ line-splitting or justification or anything. }
var
{$IFDEF PATCH_I18N}
{$ELSE PATCH_I18N}
	pline: PChar;
{$ENDIF PATCH_I18N}
	MyText: PSDL_Surface;
begin
{$IFDEF PATCH_I18N}
	MyText := I18N_TTF_RenderText( game_font , msg , Color );
{$ELSE PATCH_I18N}
	pline := QuickPCopy( msg );
	MyText := TTF_RenderText_Solid( game_font , pline , Color );
{$IFDEF Unix}
	if MyText <> Nil then SDL_SetColorKey( MyText , SDL_SRCCOLORKEY , SDL_MapRGB( MyText^.Format , 0 , 0, 0 ) );
{$ENDIF}
  {$IF DEFINED(PATCH_I18N) or DEFINED(PATCH_GH) or DEFINED(WITH_TENC)}
	sysutils.StrDispose( pline );
  {$ELSE}
	Dispose( pline );
  {$ENDIF}
{$ENDIF PATCH_I18N}
	SDL_BlitSurface( MyText , Nil , Game_Screen , @MyDest );
	SDL_FreeSurface( MyText );
end;

Procedure QuickTinyText( const msg: String; MyDest: TSDL_Rect; Color: TSDL_Color );
	{ Quickly draw some text to the screen, without worrying about }
	{ line-splitting or justification or anything. }
var
{$IFDEF PATCH_I18N}
{$ELSE PATCH_I18N}
	pline: PChar;
{$ENDIF PATCH_I18N}
	MyText: PSDL_Surface;
begin
{$IFDEF PATCH_I18N}
	MyText := I18N_TTF_RenderText( info_font , msg , Color );
{$ELSE PATCH_I18N}
	pline := QuickPCopy( msg );
	MyText := TTF_RenderText_Solid( info_font , pline , Color );
  {$IF DEFINED(PATCH_I18N) or DEFINED(PATCH_GH) or DEFINED(WITH_TENC)}
	sysutils.StrDispose( pline );
  {$ELSE}
	Dispose( pline );
  {$ENDIF}
{$ENDIF PATCH_I18N}
	MyDest.X := MyDest.X - ( MyText^.W div 2 );
	SDL_BlitSurface( MyText , Nil , Game_Screen , @MyDest );
	SDL_FreeSurface( MyText );
end;

Procedure CMessage( const msg: String; Z: TSDL_Rect; C: TSDL_Color );
	{ Print a message to the screen, centered in the requested rect. }
	{ Clear the specified zone before doing so. }
var
	MyText: PSDL_Surface;
	MyDest: TSDL_Rect;
begin
	MyText := PrettyPrint( msg , Z.W , C , True, game_font );
	if MyText <> Nil then begin
		MyDest := Z;
		MyDest.Y := MyDest.Y + ( Z.H - MyText^.H ) div 2;
		SDL_SetClipRect( Game_Screen , @Z );
		SDL_BlitSurface( MyText , Nil , Game_Screen , @MyDest );
		SDL_FreeSurface( MyText );
		SDL_SetClipRect( Game_Screen , Nil );
	end;
end;


Procedure GameMSG( const msg: string; Z: TSDL_Rect; var C: TSDL_Color; MyFont: PTTF_FONT );
	{ Print a line-justified message in the requested screen zone. }
	{ Clear the specified zone before doing so. }
var
	MyText: PSDL_Surface;
begin
	{ClrZone( Z );}
{$IFDEF PATCH_JPSSDL}
	Z.X := Z.X + Pad_Left;
	Z.Y := Z.Y + Pad_Top;
	Z.W := Z.W - Pad_Left;
	Z.H := Z.H - Pad_Top;
{$ENDIF PATCH_JPSSDL}
	MyText := PrettyPrint( msg , Z.W , C , False, MyFont );
	if MyText <> Nil then begin
		SDL_SetClipRect( Game_Screen , @Z );
		SDL_BlitSurface( MyText , Nil , Game_Screen , @Z );
		SDL_FreeSurface( MyText );
		SDL_SetClipRect( Game_Screen , Nil );
	end;
end;

Procedure GameMSG( const msg: string; Z: TSDL_Rect; var C: TSDL_Color );
	{ Call the above procedure with the default font. }
begin
	GameMSG( msg, Z, C, game_font );
end;

{$IFDEF PATCH_GH}
{$ELSE PATCH_GH}
Function DirKey( ReDrawer: RedrawProcedureType ): Integer;
	{ Get a direction selection from the user. If a standard direction }
	{ key was selected, return its direction (0 is East, increase }
	{ clockwise). See Locale.pp for details. }
	{ Return -1 if no good direction was chosen. }
var
	K: Char;
	DK: Integer;
begin
	DK := -2;
	repeat
		K := RPGKey;
		if K = KeyMap[ KMC_East ].KCode then begin
			DK := 0;
		end else if K = KeyMap[ KMC_SouthEast ].KCode then begin
			DK := 1;
		end else if K = KeyMap[ KMC_South ].KCode then begin
			DK := 2;
		end else if K = KeyMap[ KMC_SouthWest ].KCode then begin
			DK := 3;
		end else if K = KeyMap[ KMC_West ].KCode then begin
			DK := 4;
		end else if K = KeyMap[ KMC_NorthWest ].KCode then begin
			DK := 5;
		end else if K = KeyMap[ KMC_North ].KCode then begin
			DK := 6;
		end else if K = KeyMap[ KMC_NorthEast ].KCode then begin
			DK := 7;
		end else if K = RPK_TimeEvent then begin
			ReDrawer;
			GHFlip();
		end else begin
			DK := -1;
		end;
	until DK <> -2;
	DirKey := DK;
end;
{$ENDIF PATCH_GH}

Procedure EndOfGameMoreKey;
	{ The end of the game has been reached. Wait for the user to }
	{ press either the space bar or the ESC key. }
var
	A: Char;
begin
{$IFDEF PATCH_GH}
  {$IFDEF PATCH_I18N}
	DialogMSG( I18N_MsgString('EndOfGameMoreKey','Hit space bar') );
  {$ELSE PATCH_I18N}
	DialogMSG( '[Hit space bar.]' );
  {$ENDIF PATCH_I18N}
{$ENDIF PATCH_GH}

	{ Keep reading keypresses until either a space or an ESC/Backspace is found. }
	repeat
		A := RPGKey;
{$IFDEF PATCH_GH}
		if A = RPK_RightButton then begin
			A := #8;
		end;
{$ENDIF PATCH_GH}
	until ( A = ' ' ) or ( A = #27 ) or ( A = #8 );
end;

Procedure RedrawConsole;
	{ Redraw the console. Yay! }
var
	SL: SAttPtr;
	MyDest: TSDL_Rect;
	NumLines,LineNum: Integer;
{$IFDEF PATCH_GH}
	CH_tmp: SAttPtr;
	msg: String;
	NextWord: String;
	TheLine: String;
{$ENDIF PATCH_GH}
begin
	SDL_SetClipRect( Game_Screen , @ZONE_Dialog );

	MyDest := ZONE_Dialog;
	NumLines := ( ZONE_Dialog.H div TTF_FontLineSkip( game_font ) ) + 1;
{$IFDEF PATCH_GH}
	CH_tmp := NIL;

	SL := RetrieveSAtt( Console_History , NumSAtts( Console_History ) - NumLines + 1 );
	if ( NIL = SL ) then begin
		SL := Console_History;
	end;
	while ( NIL <> SL ) do begin
		msg := SL^.Info;
		TheLine := '';
		NextWord := '';
		repeat
			GetNextLine( TheLine , msg , NextWord , ZONE_Dialog.W , game_font );
			if ( '' <> TheLine ) then begin
				StoreSAtt( CH_tmp , TheLine );
			end;
			TheLine := NextWord;
		until ( '' = TheLine );
		SL := SL^.Next;
	end;

	LineNum := NumLines;
	SL := RetrieveSAtt( CH_tmp , NumSAtts( CH_tmp ) - NumLines + 1 );
	if ( NIL = SL ) then begin
		SL := CH_tmp;
		LineNum := NumSAtts( CH_tmp );
	end;
	while ( NIL <> SL ) do begin
		msg := SL^.Info;

		{ Set the coords for this line. }
		MyDest.X := ZONE_Dialog.X;
		MyDest.Y := ZONE_Dialog.Y + ZONE_Dialog.H - LineNum * TTF_FontLineSkip( game_font );

		{ Output the line. }
		QuickText( SL^.Info , MyDest , InfoGreen );

		Dec( LineNum );
		SL := SL^.Next;
	end;

	DisposeSAtt( CH_tmp );
{$ELSE PATCH_GH}
	LineNum := NumLines;
	SL := RetrieveSAtt( Console_History , NumSAtts( Console_History ) - NumLines + 1 );
	if SL = Nil then begin
		SL := Console_History;
		LineNum := NumSAtts( Console_History );
	end;

	while LineNum > 0 do begin
		{ Set the coords for this line. }
		MyDest.X := ZONE_Dialog.X;
		MyDest.Y := ZONE_Dialog.Y + ZONE_Dialog.H - LineNum * TTF_FontLineSkip( game_font );

		{ Output the line. }
		QuickText( SL^.Info , MyDest , InfoGreen );

		Dec( LineNum );
		SL := SL^.Next;
	end;
{$ENDIF PATCH_GH}

	{ Restore the clip zone to the full screen. }
	SDL_SetClipRect( Game_Screen , Nil );
end;


Procedure DialogMSG( msg: string );
	{ Print a message in the scrolling dialog box, }
	{ then store the line in Console_History. }
	{ Don't worry about screen output since the console will be redrawn the next time }
	{ the screen updates. }
var
{$IFDEF PATCH_GH}
{$ELSE PATCH_GH}
	NextWord: String;
	THELine: String;	{The line under construction.}
{$ENDIF PATCH_GH}
	SA: SAttPtr;
begin
{$IFDEF DEBUG}
	if DEBUG_TraceMacro then begin
		ErrorMessage_fork( 'TRACE: DialogMSG() "' + msg + '"');
	end;
{$ENDIF DEBUG}
	{ CLean up the message a bit. }
	DeleteWhiteSpace( msg );
	if msg = '' then Exit;
	msg := '> ' + Msg;

{$IFDEF PATCH_GH}
	if ( Console_History_Length <= NumSAtts( Console_History ) ) then begin
		SA := Console_History;
		RemoveSAtt( Console_History , SA );
		PurgeSAtt( Console_History );
	end;
	StoreSAtt( Console_History , msg );
{$ELSE PATCH_GH}
	{THELine = The first word in this iteration}
  {$IFDEF PATCH_I18N}
	THELine := ' ';
  {$ELSE PATCH_I18N}
	THELine := ExtractWord( msg );
  {$ENDIF PATCH_I18N}
	NextWord := '';

	{Start the main processing loop.}
	while TheLine <> '' do begin
		GetNextLine( TheLine , msg , NextWord , ZONE_Dialog.w, Game_Font );

		{ If appropriate, save the line. }
		if TheLine <> '' then begin
			if NumSAtts( Console_History ) >= Console_History_Length then begin
				SA := Console_History;
				RemoveSAtt( Console_History , SA );
  {$IFDEF PATCH_GH}
				PurgeSAtt( Console_History );
  {$ENDIF PATCH_GH}
			end;
			StoreSAtt( Console_History , TheLine );
		end;


		{ Prepare for the next iteration. }
		TheLine := NextWord;
	end; { while TheLine <> '' }
{$ENDIF PATCH_GH}
end;


Procedure ClearExtendedBorder( Dest: TSDL_Rect );
	{ Draw the inner box for border displays. }
begin
	Dest.X := Dest.X - 1;
	Dest.Y := Dest.Y - 1;
	Dest.W := Dest.W + 2;
	Dest.H := Dest.H + 2;
	SDL_FillRect( game_screen , @Dest , SDL_MapRGB( Game_Screen^.Format , 0 , 0 , 0 ) );
end;

Function GrowRect( MyRect: TSDL_Rect; GrowX,GrowY: Integer ): TSDL_Rect;
	{ Expand this rect by the requested amount, remaining centered on the }
	{ original rect. }
begin
	MyRect.x := MyRect.x - GrowX;
	MyRect.y := MyRect.y - GrowY;
	MyRect.w := MyRect.w + 2 * GrowX;
	MyRect.h := MyRect.h + 2 * GrowY;
	GrowRect := MyRect;
end;

Procedure FillRectWithSprite( MyRect: TSDL_Rect; MySprite: SensibleSpritePtr; MyFrame,OffX,OffY: Integer );
	{ Fill this area of the screen perfectly with the provided sprite. }
var
	MyDest: TSDL_Rect;
	X,Y,GridW,GridH: Integer;
begin
	GridW := MyRect.W div MySprite^.W + 1;
	GridH := MyRect.H div MySprite^.H + 1;
	SDL_SetClipRect( Game_Screen , @MyRect );

	MyRect.X := MyRect.X + (OffX mod MySprite^.W) - MySprite^.W;
	MyRect.Y := MyRect.Y + (OffY mod MySprite^.H) - MySprite^.H;

	{ Draw the backdrop. }
	for X := 0 to GridW do begin
		MyDest.X := MyRect.X + X * MySprite^.W;
		for Y := 0 to GridH do begin
			MyDest.Y := MyRect.Y + Y * MySprite^.H;
			DrawSprite( MySprite , MyDest , MyFrame );
		end;
	end;

	SDL_SetClipRect( Game_Screen , Nil );
end;

Procedure FillRectWithSprite( MyRect: TSDL_Rect; MySprite: SensibleSpritePtr; MyFrame: Integer );
	{ Do a FillRect with offset 0,0. }
begin
	FillRectWithSprite( MyRect, MySprite, MyFrame, 0, 0 );
end;

Procedure InfoBox( MyBox: TSDL_Rect );
	{ Do a box for drawing something else inside of. }
const
	tex_width = 16;
	border_width = tex_width div 2;
	half_dat = border_width div 2;
var
	MyFill,Dest: TSDL_Rect;
	X0,Y0,W32,H32,X,Y: Integer;
begin
	{ Fill the middle of the box with the backdrop. }
	MyFill := GrowRect( MyBox, 4, 4 );
	FillRectWithSprite( MyFill, Infobox_Backdrop, 0 );

	{ Expand the rect to its full dimensions, and draw the outline. }
	MyFill := GrowRect( MyBox, 8, 8 );
	DrawSprite( Infobox_Border , MyFill , 0 );

	Dest.X := MyFill.X;
	Dest.Y := MyFill.Y + MyFill.H - 8;
	DrawSprite( Infobox_Border , Dest , 4 );

	Dest.X := MyFill.X + MyFill.W - 8;
	Dest.Y := MyFill.Y;
	DrawSprite( Infobox_Border , Dest , 3 );

	Dest.X := MyFill.X + MyFill.W - 8;
	Dest.Y := MyFill.Y + MyFill.H - 8;
	DrawSprite( Infobox_Border , Dest , 5 );

	MyFill := GrowRect( MyBox, 0, 8 );
	SDL_SetClipRect( Game_Screen , @MyFill );
	for X := 0 to ( MyFill.W div 8 + 1 ) do begin
		Dest.X := MyFill.X + X * 8;
		Dest.Y := MyFill.Y;
		DrawSprite( Infobox_Border , Dest , 1 );
		Dest.Y := MyFill.Y + MyFill.H - 8;
		DrawSprite( Infobox_Border , Dest , 1 );
	end;
	MyFill := GrowRect( MyBox, 8, 0 );
	SDL_SetClipRect( Game_Screen , @MyFill );
	for Y := 0 to ( MyFill.H div 8 + 1 ) do begin
		Dest.Y := MyFill.Y + Y * 8;
		Dest.X := MyFill.X;
		DrawSprite( Infobox_Border , Dest , 2 );
		Dest.X := MyFill.X + MyFill.W - 8;
		DrawSprite( Infobox_Border , Dest , 2 );
	end;
	SDL_SetClipRect( Game_Screen , Nil );
end;


{$IFDEF PATCH_GH}
Function GetStringFromUser( const Prompt: String; ReDrawer: RedrawProcedureType; const Init_text: String ): String;
var
	dummy: Boolean;
begin
	GetStringFromUser := GetStringFromUser( Prompt, ReDrawer, Init_text, dummy );
end;

Function GetStringFromUser(const Prompt: String; ReDrawer: RedrawProcedureType ): String;
var
	dummy: Boolean;
begin
	GetStringFromUser := GetStringFromUser( Prompt, ReDrawer, '', dummy );
end;
{$ENDIF PATCH_GH}

{$IFDEF PATCH_GH}
Function GetStringFromUser(const Prompt: String; ReDrawer: RedrawProcedureType; const Init_text: String; var canceled: Boolean ): String;
{$ELSE PATCH_GH}
Function GetStringFromUser(const Prompt: String; ReDrawer: RedrawProcedureType ): String;
{$ENDIF PATCH_GH}
	{ Does what it says. }
const
{$IFDEF PATCH_I18N}
	MaxInputWidth = 80;
	WCLen = 16;
{$ELSE PATCH_I18N}
	AllowableCharacters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ 1234567890()-=_+,.?"*';
	MaxInputLength = 80;
{$ENDIF PATCH_I18N}
var
	A: Char;
	it: String;
	MyBigBox,MyInputBox,MyDest: TSDL_Rect;
{$IFDEF PATCH_I18N}
	Unicode: Word;
	work_dst: Array[0..WCLen] of Char;
	work_pdst: PChar;
	state: ShortInt = 0;
	mbchar_work: String = '';
{$ENDIF PATCH_I18N}
begin
	{ Initialize string. }
{$IFDEF PATCH_GH}
	it := Init_text;
	canceled := False;
{$ELSE PATCH_GH}
	it := '';
{$ENDIF PATCH_GH}

{$IFDEF WITHOUT_SDLIM}
	if ( NIL <> ReDrawer ) then begin
		ReDrawer;
		GHFlip;
	end;
	MyBigBox := ZONE_TextInputBigBox.GetRect();
	it := w32eb.EditBox( MyBigBox.X , MyBigBox.Y , MyBigBox.W , MyBigBox.H , Prompt , it , canceled );
{$ELSE WITHOUT_SDLIM}
	repeat
		MyBigBox := ZONE_TextInputBigBox.GetRect();
		MyInputBox := ZONE_TextInput.GetRect();

		{ Set up the display. }
		if ReDrawer <> Nil then ReDrawer;
		InfoBox( MyBigBox );
		{SDL_FillRect( game_screen , @MyBigBox , SDL_MapRGB( Game_Screen^.Format , BorderBlue.R , BorderBlue.G , BorderBlue.B ) );}
		SDL_FillRect( game_screen , @MyInputBox , SDL_MapRGB( Game_Screen^.Format , StdBlack.R , StdBlack.G , StdBlack.B ) );

		CMessage( Prompt , ZONE_TextInputPrompt.GetRect() , StdWhite );
		CMessage( it , MyInputBox , InfoGreen );
		MyDest.Y := MyInputBox.Y + 2;
  {$IFDEF PATCH_I18N}
		if TERMINAL_bidiRTL then begin
			MyDest.X := MyInputBox.X + ( MyInputBox.W div 2 ) - ( TextLength( Game_Font , it ) div 2 ) - Cursor_Sprite^.W;
		end else begin
			MyDest.X := MyInputBox.X + ( MyInputBox.W div 2 ) + ( TextLength( Game_Font , it ) div 2 );
		end;
  {$ELSE PATCH_I18N}
		MyDest.X := MyInputBox.X + ( MyInputBox.W div 2 ) + ( TextLength( Game_Font , it ) div 2 );
  {$ENDIF PATCH_I18N}
		DrawSprite( Cursor_Sprite , MyDest , ( Animation_Phase div 2 ) mod 4 );

		GHFlip;
  {$IFDEF PATCH_I18N}
		A := RPGKey( Unicode );

		if (0 < Unicode) then begin
			work_pdst := work_dst;
			Conv_FromUni16( @Unicode, 2, work_pdst, WCLen );
			A := EditMBCharStr( it, 127, MaxInputWidth, #0, work_pdst, state, mbchar_work );
    {$IFDEF PATCH_GH}
		end else if ( RPK_TimeEvent <> A ) and ( RPK_MouseMotion <> A ) then begin
    {$ELSE PATCH_GH}
		end else if RPK_TimeEvent <> A then begin
    {$ENDIF PATCH_GH}
			A := EditMBCharStr( it, 127, MaxInputWidth, A, NIL, state, mbchar_work );
		end;
  {$ELSE PATCH_I18N}
		A := RPGKey;

		if ( A = #8 ) and ( Length( it ) > 0 ) then begin
			it := Copy( it , 1 , Length( it ) - 1 );
		end else if ( Pos( A , AllowableCharacters ) > 0 ) and ( Length( it ) < MaxInputLength ) then begin
			it := it + A;
		end;
  {$ENDIF PATCH_I18N}
  {$IFDEF PATCH_GH}
	until ( A = #10 ) or ( A = #13 ) or ( A = #27 );
  {$ELSE PATCH_GH}
	until ( A = #13 ) or ( A = #27 );
  {$ENDIF PATCH_GH}
  {$IFDEF PATCH_GH}
	if (#27 = A) then begin
		it := '';
		canceled := True;
	end;
  {$ENDIF PATCH_GH}
{$ENDIF WITHOUT_SDLIM}

	GetStringFromUser := it;
end;

Function MsgString(const MsgLabel: String ): String;
	{ Return the standard message string which has the requested }
	{ label. }
begin
	MsgString := SAttValue( Text_Messages , MsgLabel );
end;

{$IFDEF PATCH_GH}
{$ELSE PATCH_GH}
Function MoreHighFirstLine( LList: SAttPtr ): Integer;
	{ Determine the highest possible FirstLine value. }
var
	it: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = LList) then Exit(0);
{$ENDIF PATCH_GH}
	it := NumSAtts( LList ) - ( ZONE_MoreText.H  div  TTF_FontLineSkip( game_font ) ) + 1;
	if it < 1 then it := 1;
	MoreHighFirstLine := it;
end;
{$ENDIF PATCH_GH}

{$IFDEF PATCH_GH}
Procedure MoreTextRedraw;
var
	MyPromptZone,MyTextZone: TSDL_Rect;
begin
	MyTextZone := ZONE_MoreText.GetRect();
	MyPromptZone := ZONE_MorePrompt.GetRect();

	InfoBox( MyTextZone );
	InfoBox( MyPromptZone );
	CMessage( I18N_MsgString( 'MoreText' , 'Prompt' ) , MyPromptZone , InfoGreen );
end;
{$ENDIF PATCH_GH}

Procedure MoreText( LList: SAttPtr; FirstLine: Integer; ReDrawer: RedrawProcedureType );
	{ Browse this text file across the majority of the screen. }
	{ Clear the screen upon exiting, though restoration of the }
	{ previous display is someone else's responsibility. }
{$IFDEF PATCH_GH}
var
	MT_tmp: SAttPtr;
{$ENDIF PATCH_GH}

	Procedure DisplayTextHere( const MyZone: TSDL_Rect );
	var
		T: Integer;
		MyDest: TSDL_Rect;
		MyImage: PSDL_Surface;
		CLine: SAttPtr;	{ Current Line }
{$IFDEF PATCH_I18N}
{$ELSE PATCH_I18N}
		PLine: PChar;
{$ENDIF PATCH_I18N}
	begin
		{ Set the clip area. }
		SDL_SetClipRect( Game_Screen , @MyZone );
		MyDest := MyZone;
{$IFDEF PATCH_JPSSDL}
		MyDest.X := MyDest.X + Pad_Left;
		MyDest.W := MyDest.W - Pad_Left;
{$ENDIF PATCH_JPSSDL}

		{ Error check. }
{$IFDEF PATCH_GH}
		if ( -1 = FirstLine ) then begin
			FirstLine := NumSAtts( MT_tmp ) - ( ZONE_MoreText.H  div  TTF_FontLineSkip( game_font ) ) + 1;
		end;
		if ( FirstLine < 1 ) then begin
			FirstLine := 1
		end else if ( NumSAtts( MT_tmp ) < FirstLine ) then begin
			FirstLine := NumSAtts( MT_tmp );
		end;
{$ELSE PATCH_GH}
		if FirstLine < 1 then FirstLine := 1
		else if FirstLine > MoreHighFirstLine( LList ) then FirstLine := MoreHighFirstLine( LList );
{$ENDIF PATCH_GH}

{$IFDEF PATCH_GH}
		CLine := RetrieveSAtt( MT_tmp , FirstLine );
{$ELSE PATCH_GH}
		CLine := RetrieveSATt( LList , FirstLine );
{$ENDIF PATCH_GH}
		for t := 1 to ( MyZone.H  div  TTF_FontLineSkip( game_font ) ) do begin
			if CLine <> Nil then begin
{$IFDEF PATCH_I18N}
				MyImage := I18N_TTF_RenderText( game_font , CLine^.Info , NeutralGrey );
				if TERMINAL_bidiRTL then begin
					MyDest.X := MyDest.X + MyDest.W - MyImage^.W;
				end;
{$ELSE PATCH_I18N}
				pline := QuickPCopy( CLine^.Info );
				MyImage := TTF_RenderText_Solid( game_font , pline , NeutralGrey );
  {$IF DEFINED(PATCH_I18N) or DEFINED(PATCH_GH) or DEFINED(WITH_TENC)}
				sysutils.StrDispose( pline );
  {$ELSE}
				Dispose( pline );
  {$ENDIF}
{$IFDEF Unix}
				SDL_SetColorKey( MyImage , SDL_SRCCOLORKEY , SDL_MapRGB( MyImage^.Format , 0 , 0, 0 ) );
{$ENDIF}
{$ENDIF PATCH_I18N}

				SDL_BlitSurface( MyImage , Nil , Game_Screen , @MyDest );
				SDL_FreeSurface( MyImage );
				MyDest.Y := MyDest.Y + TTF_FontLineSkip( game_font );
				CLine := CLine^.Next;
			end;
		end;

		if (Animation_Phase div 10 mod 2) = 1 then begin
			if FirstLine > 1 then begin
				MyDest.X := MyZone.X + MyZone.W - 16;
				MyDest.Y := MyZone.Y;
				QuickText('+', MyDest, MenuSelect );
			end;
			if CLine <> Nil then begin
				MyDest.X := MyZone.X + MyZone.W - 16;
				MyDest.Y := MyZone.Y + MyZone.H - TTF_FontLineSkip( game_font );
				QuickText('+', MyDest, MenuSelect );
			end;
		end;

		{ Restore the clip area. }
		SDL_SetClipRect( Game_Screen , Nil );
		GHFlip;
	end;
var
	A: Char;
	MyPromptZone,MyTextZone: TSDL_Rect;
{$IFDEF PATCH_GH}
	RPM: RPGMenuPtr;
	t, t_max: Integer;
	CLine: SAttPtr;	{ Current Line }
	msg: String;
	NextWord: String;
	TheLine: String;
	cont: Boolean;
	stored: Boolean;
{$ENDIF PATCH_GH}
begin
{$IFDEF PATCH_GH}
	if SDL_Show_MenuScrollbar then begin
		RPM := CreateRPGMenu( NeutralGrey , MenuSelect , ZONE_MoreText );
		CLine := RetrieveSAtt( LList , 1 );
		t_max := NumSAtts( LList );
		for t := 1 to t_max do begin
			msg := CLine^.Info;
			TheLine := '';
			NextWord := '';
			cont := false;
			stored := false;
			repeat
				GetNextLine( TheLine , msg , NextWord , ZONE_MoreText.W , game_font );
				if ('' <> TheLine) then begin
					if cont then begin
						AddRPGMenuItem( RPM , #$0 + TheLine , -1 );
					end else begin
						AddRPGMenuItem( RPM , TheLine , -1 );
						cont := true;
					end;
					stored := true;
				end;
				TheLine := NextWord;
			until ('' = TheLine);
			if not stored then begin
				AddRPGMenuItem( RPM , '' , -1 );
			end;
			CLine := CLine^.Next;
		end;
		if ( -1 = FirstLine ) then begin
			FirstLine := RPM^.NumItem - ( ZONE_MoreText.H  div  TTF_FontLineSkip( game_font ) ) + 1;
		end;
		if ( FirstLine < 1 ) then begin
			FirstLine := 1
		end else if ( RPM^.NumItem < FirstLine ) then begin
			FirstLine := RPM^.NumItem;
		end;
		RPM^.TopItem := FirstLine;
		SetItemByPosition( RPM, FirstLine );
		SelectMenu( RPM , @MoreTextRedraw );
		DisposeRPGMenu( RPM );
	end else begin
{$ENDIF PATCH_GH}

{$IFDEF PATCH_GH}
	MT_tmp := NIL;
	CLine := LList;
	while ( NIL <> CLine ) do begin
		msg := CLine^.Info;
		TheLine := '';
		NextWord := '';
		stored := false;
		repeat
			GetNextLine( TheLine , msg , NextWord , ZONE_MoreText.W , game_font );
			if ( '' <> TheLine ) then begin
				StoreSAtt( MT_tmp , TheLine );
				stored := true;
			end;
			TheLine := NextWord;
		until ( '' = TheLine );
		if not stored then begin
			StoreSAtt( MT_tmp , '' );
		end;
		CLine := CLine^.Next;
	end;

	{ Display the screen. }
	MyTextZone := ZONE_MoreText.GetRect();
	MyPromptZone := ZONE_MorePrompt.GetRect();
	if Redrawer <> Nil then Redrawer();
	InfoBox( MyTextZone );
	InfoBox( MyPromptZone );
	CMessage( MsgString( 'MORETEXT_Prompt' ) , MyPromptZone , InfoGreen );
	DisplayTextHere( MyTextZone );
{$ENDIF PATCH_GH}

	repeat
		{ Get input from user. }
		A := RPGKey;

		{ Possibly process this input. }
{$IFDEF PATCH_GH}
		if (KBD_NONE = A) then begin
		end else if (KeyMap[ KMC_MenuUp    ].KCode = A) then begin Dec( FirstLine );
		end else if (KeyMap[ KMC_MenuDown  ].KCode = A) then begin Inc( FirstLine );
		end else if (KeyMap[ KMC_PageUp    ].KCode = A) or (KeyMap[ KMC_ButtonWLeft  ].KCode = A) or (KeyMap[ KMC_ButtonWBack    ].KCode = A) then begin FirstLine := FirstLine - ( ZONE_MoreText.H div TTF_FontLineSkip( game_font ) ); DisplayTextHere( MyTextZone );
		end else if (KeyMap[ KMC_PageDown  ].KCode = A) or (KeyMap[ KMC_ButtonWRight ].KCode = A) or (KeyMap[ KMC_ButtonWForward ].KCode = A) then begin FirstLine := FirstLine + ( ZONE_MoreText.H div TTF_FontLineSkip( game_font ) ); DisplayTextHere( MyTextZone );
  {$IFDEF PATCH_JPSSDL}
		end else if (KeyMap[ KMC_MenuLeft  ].KCode = A) or (KeyMap[ KMC_ScrollUp     ].KCode = A) or (KeyMap[ KMC_ButtonWUp      ].KCode = A) then begin FirstLine := FirstLine - ( ZONE_MoreText.H div TTF_FontLineSkip( game_font ) ); DisplayTextHere( MyTextZone );
		end else if (KeyMap[ KMC_MenuRight ].KCode = A) or (KeyMap[ KMC_ScrollDown   ].KCode = A) or (KeyMap[ KMC_ButtonWDown    ].KCode = A) then begin FirstLine := FirstLine + ( ZONE_MoreText.H div TTF_FontLineSkip( game_font ) ); DisplayTextHere( MyTextZone );
  {$ENDIF PATCH_JPSSDL}
		end else if A = RPK_RightButton then begin A := #8;
		end else if A = RPK_TimeEvent then begin
			MyTextZone := ZONE_MoreText.GetRect();
			MyPromptZone := ZONE_MorePrompt.GetRect();
			if Redrawer <> Nil then Redrawer();
			InfoBox( MyTextZone );
			InfoBox( MyPromptZone );
			CMessage( MsgString( 'MORETEXT_Prompt' ) , MyPromptZone , InfoGreen );

			{ Display the screen. }
			DisplayTextHere( MyTextZone );
		end;

	until (KBD_ESC = A) or ('Q' = A) or (#8 = A);
{$ELSE PATCH_GH}
		if A = RPK_Down then begin
			Inc( FirstLine );
		end else if A = RPK_Up then begin
			Dec( FirstLine );
  {$IFDEF PATCH_JPSSDL}
		end else if A = RPK_Left then begin
			FirstLine := FirstLine - ( ZONE_MoreText.H  div  TTF_FontLineSkip( game_font ) );
			DisplayTextHere( MyTextZone );
		end else if A = RPK_Right then begin
			FirstLine := FirstLine + ( ZONE_MoreText.H  div  TTF_FontLineSkip( game_font ) );
			DisplayTextHere( MyTextZone );
  {$ENDIF PATCH_JPSSDL}
		end else if A = RPK_TimeEvent then begin
			MyTextZone := ZONE_MoreText.GetRect();
			MyPromptZone := ZONE_MorePrompt.GetRect();
			if Redrawer <> Nil then Redrawer();
			InfoBox( MyTextZone );
			InfoBox( MyPromptZone );
			CMessage( MsgString( 'MORETEXT_Prompt' ) , MyPromptZone , InfoGreen );

			{ Display the screen. }
			DisplayTextHere( MyTextZone );
		end;

	until ( A = #27 ) or ( A = 'Q' ) or ( A = #8 );
{$ENDIF PATCH_GH}
{$IFDEF PATCH_GH}
	DisposeSAtt( MT_tmp );
{$ENDIF PATCH_GH}
{$IFDEF PATCH_GH}
	end;
{$ENDIF PATCH_GH}
end;

Procedure DrawBPBorder;
	{ Draw borders for the backpack display. }
var
	MyRect: TSDL_Rect;
begin
	MyRect := ZONE_BPTotal.GetRect();
	ClearExtendedBorder( MyRect );
	SDL_FillRect( game_screen , @MyRect , SDL_MapRGB( Game_Screen^.Format , BorderBlue.R , BorderBlue.G , BorderBlue.B ) );
	ClearExtendedBorder( ZONE_EqpMenu.GetRect() );
	ClearExtendedBorder( ZONE_InvMenu.GetRect() );
	ClearExtendedBorder( ZONE_BPHeader.GetRect() );
	ClearExtendedBorder( ZONE_BPInstructions.GetRect() );
	ClearExtendedBorder( ZONE_BPInfo.GetRect() );
	FillRectWithSprite( ZONE_BPInfo.GetRect(), Infobox_Backdrop, 0 );
end;


Procedure DrawCharGenBorder;
	{ Draw borders for the character generator. }
begin
	SDL_FillRect( game_screen , Nil , SDL_MapRGB( Game_Screen^.Format , 47 , 64 , 91 ) );
	InfoBox( ZONE_CharGenChar.GetRect() );
	InfoBox( ZONE_CharGenMenu.GetRect() );
	InfoBox( ZONE_CharGenDesc.GetRect() );
	InfoBox( ZONE_CharGenPrompt.GetRect() );
	InfoBox( ZONE_CharGenCaption.GetRect() );
end;

Procedure SetupCombatDisplay;
	{ Draw the combat background. }
begin
{	SDL_FillRect( game_screen , Nil , SDL_MapRGB( Game_Screen^.Format , BorderBlue.R , BorderBlue.G , BorderBlue.B ) );
	ClearExtendedBorder( ZONE_Map );
	ClearExtendedBorder( ZONE_Info );
	ClearExtendedBorder( ZONE_Menu );
	ClearExtendedBorder( ZONE_Dialog );
	ClearExtendedBorder( ZONE_Clock );}
end;

Procedure SetupHQDisplay;
begin
end;
Procedure SetupFactoryDisplay;
begin
end;

Procedure SetupYesNoDisplay;
	{ Draw an outline around the YesNoMenu display area. }
begin
	ClearExtendedBorder( ZONE_YesNoTotal.GetRect() );
	SDL_FillRect( game_screen , @ZONE_YesNoTotal , SDL_MapRGB( Game_Screen^.Format , BorderBlue.R , BorderBlue.G , BorderBlue.B ) );
	ClearExtendedBorder( ZONE_YesNoPrompt.GetRect() );
	ClearExtendedBorder( ZONE_YesNoMenu.GetRect() );
end;

Procedure SetupMemoDisplay;
	{ Draw an outline around the memo display. Fortunately, that's the same region as the }
	{ YesNo display. }
begin
	InfoBox( ZONE_MemoTotal.GetRect() );
end;

Procedure SetupInteractDisplay( TeamColor: TSDL_Color );
	{ Draw the display for the interaction interface. }
var
	MyDest: TSDL_Rect;
begin
	MyDest := ZONE_InteractTotal.GetRect();
	ClearExtendedBorder( MyDest );
	SDL_FillRect( game_screen , @MyDest , SDL_MapRGB( Game_Screen^.Format , TeamColor.R , TeamColor.G , TeamColor.B ) );
	ClearExtendedBorder( ZONE_InteractStatus.GetRect() );
	ClearExtendedBorder( ZONE_InteractMsg.GetRect() );
	ClearExtendedBorder( ZONE_InteractMenu.GetRect() );
	ClearExtendedBorder( ZONE_InteractPhoto.GetRect() );
	ClearExtendedBorder( ZONE_InteractInfo.GetRect() );
{$IFDEF PATCH_GH}
	ClearExtendedBorder( ZONE_InteractPhotoFrame.GetRect() );
{$ENDIF PATCH_GH}
end;


{$IFDEF PATCH_GH}
Procedure CheckSleepGranularity;
var
	i, j, t, Count, Granularity: UInt32;
	sGranularity: Single;
begin
	if 0 > SleepGranularity then begin
		Count := 40;
		for j := 0 to 1 do begin
			t := 0;
			i := 0;
			Last_Clock_Update := SDL_GetTicks;
			while (i < Count) do begin
				SDL_Delay( i );
				t := t + i;
				i := i + 2;
			end;
			Last_Clock_Update := SDL_GetTicks - Last_Clock_Update;
			sGranularity := (Last_Clock_Update - t);
			sGranularity := sGranularity / Count;
			Granularity := (Last_Clock_Update - t) div Count;
		end;

		if -1 > SleepGranularity then begin
			ErrorMessage( 'Granularity of SDL_Delay is ' + IntToStr(trunc(sGranularity)) + '.' + IntToStr(trunc(sGranularity * 10.0) mod 10) + IntToStr(trunc(sGranularity * 100.0) mod 10) + IntToStr(trunc(sGranularity * 1000.0) mod 10) + ' [ms] .' );
		end;
		SleepGranularity := Granularity;
		Last_Clock_Update := 0;
	end;
end;
{$ENDIF PATCH_GH}

{$IFDEF PATCH_I18N}
Function SearchAndOpenFont( const FontInfo: PFontSearchNameDesc; arg_ptsize: integer ): PTTF_Font;
const
	TmpLen = 255;
var
	i, j: Integer;
	FontFile: String;
	tmp: array[0..TmpLen] of Char;
	ptsize: Integer;
begin
	SearchAndOpenFont := NIL;
	for j := 0 to (MaxFontSearchNameNum-1) do begin
		for i := 1 to MaxFontSearchDirNum do begin
			if 0 < Length(FontSearchDir[i]) then begin
				FontFile := FontSearchDir[i] + OS_Dir_Separator + FontInfo[j].FontFile;
			end else begin
				FontFile := FontInfo[j].FontFile;
			end;
			ptsize := arg_ptsize;
			if ptsize <= 0 then begin
				ptsize := FontInfo[j].FontSize;
			end;
			if 0 < Length(FontInfo[j].FontFile) then begin
				StrPCopy( tmp, FontFile );
  {$IFDEF DEBUG}
    {$IFDEF PATCH_GH}
				ErrorMessage_fork( '[Name:' + BStr(j) + ';Dir:' + BStr(i) + ']=' + FontFile );
    {$ENDIF PATCH_GH}
  {$ENDIF DEBUG}
				SearchAndOpenFont := TTF_OpenFontIndex( @tmp, ptsize, FontInfo[j].FontFace );
				if (NIL <> SearchAndOpenFont) then
					break;
			end;
		end;
		if (NIL <> SearchAndOpenFont) then
			break;
	end;
end;
{$ENDIF PATCH_I18N}


Procedure SetupWizardDisplay();
	{ This procedure will set the Wizard display decorations and resize all }
	{ the relevant game zones. Yay? }
var
	MyRect: TSDL_Rect;
	Procedure AddHorizontalDivider( Y: Integer );
		{ Add a divider to the Wizard UI. }
	var
		Dest: TSDL_Rect;
		X: Integer;
	begin
		Dest.X := MyRect.X-8;
		Dest.Y := Y;
		DrawSprite( Infobox_Border, Dest, 6 );
		Dest.X := MyRect.X + MyRect.W;
		DrawSprite( Infobox_Border, Dest, 7 );

		Dest := GrowRect( MyRect, 0, 8 );
		SDL_SetClipRect( Game_Screen , @Dest );
		Dest.Y := Y;
		for X := 0 to ( MyRect.W div 8 + 1 ) do begin
			Dest.X := MyRect.X + X * 8;
			DrawSprite( Infobox_Border , Dest , 1 );
		end;
		SDL_SetClipRect( Game_Screen , Nil );
	end;
	Procedure AddVerticalDivider( X: Integer );
		{ Add a divider to the Wizard UI. }
	var
		Dest: TSDL_Rect;
		Y: Integer;
	begin
		Dest.X := X;
		Dest.Y := MyRect.Y-8;
		DrawSprite( Infobox_Border, Dest, 8 );
		Dest.Y := MyRect.Y + MyRect.H;
		DrawSprite( Infobox_Border, Dest, 9 );

		Dest := GrowRect( MyRect, 8, 0 );
		SDL_SetClipRect( Game_Screen , @Dest );
		Dest.X := X;
		for Y := 0 to ( MyRect.H div 8 + 1 ) do begin
			Dest.Y := MyRect.Y + Y * 8;
			DrawSprite( Infobox_Border , Dest , 2 );
		end;
		SDL_SetClipRect( Game_Screen , Nil );
	end;
{$IFDEF PATCH_GH}
var
	DA_H: Integer;
{$ENDIF PATCH_GH}
begin
{$IFDEF PATCH_GH}
	if ( 0 < SDL_Change_DialogHeight ) then begin
		DA_H := TTF_FontLineSkip( Game_Font ) * SDL_Change_DialogHeight + 1;
		if ( DA_H < Dialog_Area_Height ) then begin
			DA_H := Dialog_Area_Height;
		end;
	end else begin
		DA_H := Dialog_Area_Height;
	end;

	MyRect.X := Game_Screen^.w div 2 - 390;
	MyRect.Y := Game_Screen^.h - DA_H - 10;
	MyRect.w := 780;
	MyRect.h := DA_H;

	InfoBox( MyRect );

	ZONE_Dialog.x := MyRect.X + Right_Column_Width + 8;
	ZONE_Dialog.y := MyRect.Y;
	ZONE_Dialog.w := MyRect.W - Right_Column_Width - 8;
	ZONE_Dialog.h := DA_H;
{$ELSE PATCH_GH}
	MyRect.X := Game_Screen^.w div 2 - 390;
	MyRect.Y := Game_Screen^.h - Dialog_Area_Height - 10;
	MyRect.w := 780;
	MyRect.h := Dialog_Area_Height;

	InfoBox( MyRect );

	ZONE_Dialog.x := MyRect.X + Right_Column_Width + 8;
	ZONE_Dialog.y := MyRect.Y;
	ZONE_Dialog.w := MyRect.W - Right_Column_Width - 8;
	ZONE_Dialog.h := Dialog_Area_Height;
{$ENDIF PATCH_GH}

	ZONE_Map.X := 0;
	ZONE_Map.Y := 0;
	ZONE_Map.W := Game_Screen^.W;
	ZONE_Map.H := Game_Screen^.H;

	ZONE_Clock.x := MyRect.X;
	ZONE_Clock.y := Game_Screen^.h - 28;

	ZONE_PCInfo.X := MyRect.X;
	ZONE_PCInfo.Y := MyRect.Y;

	AddVerticalDivider( ZONE_Dialog.x - 8 );

{	AddDivider( 160 );
	AddDivider( Game_Screen^.H - 36 );

	if menus > 0 then begin
		AddDivider( ZONE_Menu.Y + ZONE_Menu.H );
		ZONE_Dialog.y := ZONE_Menu.Y + ZONE_Menu.H + 8;
		if menus > 1 then begin
			AddDivider( ZONE_Menu1.Y + ZONE_Menu1.H );
		end;
	end else begin
		ZONE_Dialog.y := 168;
	end;
	ZONE_Dialog.h := Game_Screen^.H - ZONE_Dialog.y - 36;
}
end;

{$IFDEF PATCH_GH}
Procedure Reset_DynamicRect;
	Procedure Set_org( DR: DynamicRectPtr );
	begin
		DR^.dx_org := DR^.dx;
		DR^.dy_org := DR^.dy;
		DR^.w_org := DR^.w;
		DR^.h_org := DR^.h;
	end;
begin
	Set_org( @ZONE_TitleScreenMenu );
	Set_org( @ZONE_TitleScreenLogo );
	Set_org( @ZONE_TitleScreenVersion );
	Set_org( @ZONE_TargetInfo );
	Set_org( @ZONE_TargetDistance );
	Set_org( @ZONE_CharGenChar );
	Set_org( @ZONE_CharGenMenu );
	Set_org( @ZONE_CharGenCaption );
	Set_org( @ZONE_CharGenDesc );
	Set_org( @ZONE_CharGenPrompt );
	Set_org( @ZONE_CharGenHint );
	Set_org( @ZONE_CharViewChar );
	Set_org( @ZONE_CharViewMenu );
	Set_org( @ZONE_CharViewCaption );
	Set_org( @ZONE_CharViewDesc );
	Set_org( @ZONE_HQPilots );
	Set_org( @ZONE_HQMecha );
	Set_org( @ZONE_HQMenu );
	Set_org( @ZONE_HQMenu2 );
	Set_org( @ZONE_HQMenu3 );
	Set_org( @ZONE_TextInputPrompt );
	Set_org( @ZONE_TextInput );
	Set_org( @ZONE_TextInputBigBox );
	Set_org( @ZONE_PhoneInstructions );
	Set_org( @ZONE_InteractStatus );
	Set_org( @ZONE_InteractMsg );
	Set_org( @ZONE_InteractMenu );
	Set_org( @ZONE_InteractPhoto );
	Set_org( @ZONE_InteractInfo );
	Set_org( @ZONE_InteractTotal );
{$IFDEF PATCH_GH}
	Set_org( @ZONE_InteractPhotoFrame );
{$ENDIF PATCH_GH}
	Set_org( @ZONE_Menu );
	Set_org( @ZONE_Menu1 );
	Set_org( @ZONE_Menu2 );
{$IFDEF PATCH_GH}
	Set_org( @ZONE_GameOptionMenu );
	Set_org( @ZONE_MenuRight );
	Set_org( @ZONE_Menu11Right );
	Set_org( @ZONE_Menu12Right );
	Set_org( @ZONE_Menu1Right );
	Set_org( @ZONE_Menu2Right );
{$ENDIF PATCH_GH}
	Set_org( @ZONE_MemoText );
	Set_org( @ZONE_MemoMenu );
	Set_org( @ZONE_MemoTotal );
	Set_org( @ZONE_CenterMenu );
	Set_org( @ZONE_FHQTitle );
	Set_org( @ZONE_FHQMenu );
	Set_org( @ZONE_FHQInfo );
	Set_org( @ZONE_FHQMenu1 );
	Set_org( @ZONE_FHQMenu2 );
	Set_org( @ZONE_BPTotal );
	Set_org( @ZONE_BPHeader );
	Set_org( @ZONE_EqpMenu );
	Set_org( @ZONE_InvMenu );
	Set_org( @ZONE_BPInstructions );
	Set_org( @ZONE_BPInfo );
	Set_org( @ZONE_ShopNPCName );
	Set_org( @ZONE_ShopNPCPortrait );
	Set_org( @ZONE_ShopText );
	Set_org( @ZONE_ShopPCName );
	Set_org( @ZONE_ShopPCPortrait );
	Set_org( @ZONE_ShopMenu );
	Set_org( @ZONE_ShopInfo );
	Set_org( @ZONE_ShopCash );
	Set_org( @ZONE_ShopTop );
	Set_org( @ZONE_ShopBottom );
	Set_org( @ZONE_MoreText );
	Set_org( @ZONE_MorePrompt );
	Set_org( @ZONE_YesNoTotal );
	Set_org( @ZONE_YesNoPrompt );
	Set_org( @ZONE_YesNoMenu );
end;
{$ENDIF PATCH_GH}



initialization
begin
{$IFDEF DEBUG}
	ErrorMessage_fork('DEBUG: sdlgfx.pp');
{$ENDIF DEBUG}

{$IFDEF PATCH_I18N}
	I18N_Width_Of_One_Character := I18N_Settings('SDLGFX_I18N_WIDTH_OF_ONE_CHARACTER','M');
{$ENDIF PATCH_I18N}

{$IFDEF PATCH_GH}
	if (ScreenSize_Width <= 0) then begin
		ScreenSize_Width := 800;
	end;
	if (ScreenSize_Height <= 0) then begin
		ScreenSize_Height := 600;
	end;
{$ENDIF PATCH_GH}

{$IF DEFINED(PATCH_AUDIO)}
	SDL_Init( SDL_INIT_VIDEO or SDL_INIT_TIMER or SDL_INIT_AUDIO );
{$ELSEIF DEFINED(PATCH)}
	SDL_Init( SDL_INIT_VIDEO or SDL_INIT_TIMER );
{$ELSE}
	SDL_Init( SDL_INIT_VIDEO or SDL_INIT_AUDIO );
{$ENDIF}

{$IFDEF PATCH_GH}
	Reset_DynamicRect;
	Recalc_DynamicRect;
{$ENDIF PATCH_GH}

{$IFDEF PATCH_GH}
	SDL_VideoModeFlag := 0;
{$ENDIF PATCH_GH}
	if DoFullScreen then begin
{$IFDEF PATCH_GH}
		SDL_VideoModeFlag := ( SDL_HWSURFACE or SDL_FULLSCREEN or SDL_DoubleBuf );
		Game_Screen := SDL_SetVideoMode(ScreenSize_Width, ScreenSize_Height, 0, SDL_VideoModeFlag );
{$ELSE PATCH_GH}
		Game_Screen := SDL_SetVideoMode(ScreenWidth, ScreenHeight, 0, SDL_HWSURFACE or SDL_FULLSCREEN or SDL_DoubleBuf );
{$ENDIF PATCH_GH}
		Mouse_Pointer := IMG_Load( Graphics_Directory + 'cosplay_pointer.png' );
		SDL_SetColorKey( Mouse_Pointer , SDL_SRCCOLORKEY or SDL_RLEACCEL , SDL_MapRGB( Mouse_Pointer^.Format , 0 , 0, 255 ) );
	end else begin
{$IFDEF PATCH_GH}
		SDL_VideoModeFlag := ( SDL_HWSURFACE or SDL_DoubleBuf or SDL_RESIZABLE );
		Game_Screen := SDL_SetVideoMode(ScreenSize_Width, ScreenSize_Height, 0, SDL_VideoModeFlag );
{$ELSE PATCH_GH}
		Game_Screen := SDL_SetVideoMode(ScreenWidth, ScreenHeight, 0, SDL_HWSURFACE or SDL_DoubleBuf or SDL_RESIZABLE );
{$ENDIF PATCH_GH}
		Mouse_Pointer := Nil;
	end;
{$IFDEF PATCH_GH}
	if (NIL = Game_Screen) then begin
		ErrorMessage('SDL_SetVideoMode() failed.');
		halt(255);
	end;
{$ENDIF PATCH_GH}

{$IFDEF PATCH_GH}
	if (NIL = Game_Screen) then begin
		ErrorMessage('SDL_ConvertSurface() failed.');
		halt(255);
	end;
{$ENDIF PATCH_GH}
{$IFDEF PATCH_GH}
	if ( ( -1 <> ScreenPos_X ) and ( -1 <> ScreenPos_Y ) ) then begin
			SDL_VERSION( wmi.version );
			if ( 1 = SDL_GetWMInfo( @wmi ) ) then begin
  {$IF DEFINED(UNIX)}
				Pointer(gX11_Lock) := wmi.x11.lock_func;
				Pointer(gX11_Unlock) := wmi.x11.unlock_func;
				gX11_Lock;
				xlib.XMoveWindow( wmi.x11.display , wmi.x11.wmwindow , ScreenPos_X , ScreenPos_Y );
				gX11_Unlock;
  {$ENDIF}
  {$IF DEFINED(WINDOWS)}
				Windows.GetWindowRect( wmi.window , @r );
				Windows.MoveWindow( wmi.window , ScreenPos_X , ScreenPos_Y , ( r.right - r.left ) , ( r.bottom - r.top ) , FALSE );
  {$ENDIF}
			end;
		end;
{$ENDIF PATCH_GH}

	SDL_EnableUNICODE( 1 );
	SDL_EnableKeyRepeat( GH_REPEAT_DELAY , GH_REPEAT_INTERVAL );


	Game_Sprites := Nil;

	Cursor_Sprite := ConfirmSprite( 'cursor.png' , '' , 8 , 16 );

	TTF_Init;
{$IFDEF PATCH_I18N}
	Game_Font := SearchAndOpenFont( @FontSearchName_Big, FontSize_Big );
	Info_Font := SearchAndOpenFont( @FontSearchName_Small, FontSize_Small );
{$ELSE PATCH_I18N}
	Game_Font := TTF_OpenFont( 'Image' + OS_Dir_Separator + 'VeraBd.ttf' , BigFontSize );
	Info_Font := TTF_OpenFont( 'Image' + OS_Dir_Separator + 'VeraMoBd.ttf' , SmallFontSize );
{$ENDIF PATCH_I18N}

	Text_Messages := LoadStringList( Standard_Message_File );
	Console_History := Nil;

{$IFDEF PATCH_GH}
	if (NIL = Game_Font) or (NIL = Info_Font) then begin
		ErrorMessage('No fonts were found.');
		halt(1);
	end;
	if (NIL = Text_Messages) then begin
		ErrorMessage('Standard_Message_File was not found.');
		halt(1);
	end;
	Attach_SmartPointer( 'Console_History: SAttPtr', @Console_History );
{$ENDIF PATCH_GH}

	SDL_WM_SetCaption( WindowName , IconName );

	Animation_Phase := 0;
{$IFDEF PATCH_GH}
	CheckSleepGranularity;
{$ENDIF PATCH_GH}
	Last_Clock_Update := 0;

	MasterColorList := LoadStringList( Data_Directory + 'sdl_colors.txt' );

	Infobox_Border := ConfirmSprite( 'sys_boxborder.png' , '', 8 , 8 );
	Infobox_Backdrop := ConfirmSprite( 'sys_boxbackdrop.png' , '', 16 , 16 );

	if Transparent_Interface then SDL_SetAlpha( Infobox_Backdrop^.Img , SDL_SRCAlpha , 224 );

{$IFDEF PATCH_GH}
	SetupWizardDisplay;
{$ENDIF PATCH_GH}

{$IFDEF PATCH_AUDIO}
	Mix_OpenAudio( MIX_DEFAULT_FREQUENCY , MIX_DEFAULT_FORMAT , 2 , 4096 );
	Mix_AllocateChannels( 16 );
	Load_SoundEffect;
{$ENDIF PATCH_AUDIO}
{	MIX_OpenAudio( MIX_DEFAULT_FREQUENCY , MIX_DEFAULT_FORMAT , MIX_CHANNELS , 4096 );
	Music_List := LoadStringList( 'music.cfg' );
	MyMusic := Nil;
}
end;

finalization
begin
{$IFDEF DEBUG}
	ErrorMessage_fork('DEBUG: sdlgfx.pp(finalization)');
{$ENDIF DEBUG}

{$IFDEF PATCH_AUDIO}
	Dispose_SoundEffect;
	Mix_CloseAudio;
{$ENDIF PATCH_AUDIO}
{	if MyMusic <> Nil then MIX_FreeMusic( MyMusic );
	MIX_CloseAudio;
	DisposeSAtt( Music_List );
}
	DisposeSpriteList( Game_Sprites );
	TTF_CloseFont( Game_Font );
	TTF_CloseFont( Info_Font );
	TTF_Quit;

	if Mouse_Pointer <> Nil then SDL_FreeSurface( Mouse_Pointer );

	SDL_FreeSurface( Game_Screen );
	SDL_Quit;

	DisposeSAtt( Text_Messages );
	DisposeSAtt( Console_History );

	DisposeSAtt( MasterColorList )

end;

end.
